from pyspedas.mms.mms_load_data import mms_load_data
from pyspedas.mms.eis.mms_eis_omni import mms_eis_omni
from pyspedas.mms.eis.mms_eis_spin_avg import mms_eis_spin_avg
from pyspedas.mms.eis.mms_eis_set_metadata import mms_eis_set_metadata
from pyspedas.mms.print_vars import print_vars
from pyspedas.mms.mms_config import CONFIG
from pyspedas import tnames


@print_vars
def mms_load_eis(trange=['2015-10-16', '2015-10-17'], probe='1', data_rate='srvy', level='l2', datatype='extof',
        varformat=None, varnames=[], get_support_data=True, suffix='', time_clip=False, no_update=False,
        available=False, notplot=False, latest_version=False, major_version=False, min_version=None, cdf_version=None, 
        spdf=False, always_prompt=False):
    """
    Load data from the Energetic Ion Spectrometer (EIS)
    
    Parameters
    ----------
        trange : list of str
            time range of interest [start time, end time] with the format
            'YYYY-MM-DD','YYYY-MM-DD'] or to specify more or less than a day 
            ['YYYY-MM-DD/hh:mm:ss','YYYY-MM-DD/hh:mm:ss']

        probe : str or list of str
            list of probes, valid values for MMS probes are ['1','2','3','4']. 

        data_rate : str or list of str
            instrument data rates for EIS include ['brst', 'srvy']. The
            default is 'srvy'.

        level : str
            indicates level of data processing. the default if no level is specified is 'l2'

        datatype : str or list of str
            Valid datatypes for EIS are: ['extof', 'phxtof', and 'electronenergy']; default is 'extof'

        get_support_data: bool
            Data with an attribute "VAR_TYPE" with a value of "support_data"
            will be loaded into tplot.  By default, only loads in data with a 
            "VAR_TYPE" attribute of "data".

        time_clip: bool
            Data will be clipped to the exact trange specified by the trange keyword.
            
        varformat: str
            The file variable formats to load into tplot.  Wildcard character
            "*" is accepted.  By default, all variables are loaded in.

        varnames: list of str
            List of variable names to load (if not specified,
            all data variables are loaded)

        suffix: str
            The tplot variable names will be given this suffix.  By default, 
            no suffix is added.

        notplot: bool
            If True, then data are returned in a hash table instead of 
            being stored in tplot variables (useful for debugging, and
            access to multidimensional data products)

        available: bool
            If True, simply return the available data files (without downloading)
            for the requested parameters

        no_update: bool
            Set this flag to preserve the original data. if not set and newer 
            data is found the existing data will be overwritten

        cdf_version: str
            Specify a specific CDF version # to load (e.g., cdf_version='4.3.0')

        min_version: str
            Specify a minimum CDF version # to load

        latest_version: bool
            Only grab the latest CDF version in the requested time interval

        major_version: bool
            Only open the latest major CDF version (e.g., X in vX.Y.Z) in the requested time interval

        always_prompt: bool
            Set this keyword to always prompt for the user's username and password;
            useful if you accidentally save an incorrect password, or if your SDC password has changed

        spdf: bool
            If True, download the data from the SPDF instead of the SDC

    Returns
    ----------
        List of tplot variables created.

    """

    tvars = mms_load_data(trange=trange, notplot=notplot, probe=probe, data_rate=data_rate, level=level, instrument='epd-eis',
            datatype=datatype, varformat=varformat, varnames=varnames, get_support_data=get_support_data, prefix='', suffix=suffix,
            time_clip=time_clip, no_update=no_update, available=available, latest_version=latest_version, 
            major_version=major_version, min_version=min_version, cdf_version=cdf_version, spdf=spdf, always_prompt=always_prompt)

    if tvars == [] or available or notplot or CONFIG['download_only']:
        return tvars

    if not isinstance(probe, list): probe = [probe]
    if not isinstance(data_rate, list): data_rate = [data_rate]
    if not isinstance(level, list): level = [level]
    if not isinstance(datatype, list): datatype = [datatype]

    # the probes will need to be strings beyond this point
    if isinstance(probe, list):
        probe = [str(p) for p in probe]

    for probe_id in probe:
        for datatype_id in datatype:
            for level_id in level:
                for data_rate_id in data_rate:
                    if datatype_id == 'electronenergy':
                        e_spin_avg_var = mms_eis_spin_avg(probe=probe_id, species='electron', datatype=datatype_id, data_rate=data_rate_id, level=level_id, suffix=suffix)
                        # create non-spin averaged omni-directional spectra
                        e_omni_spectra = mms_eis_omni(probe_id, species='electron', data_rate=data_rate_id, level=level_id, datatype=datatype_id)
                        # create spin averaged omni-directional spectra
                        e_omni_spectra_spin = mms_eis_omni(probe_id, species='electron', data_rate=data_rate_id, level=level_id, datatype=datatype_id, suffix=suffix+'_spin')
                        # add the vars to the output
                        if e_spin_avg_var is not None:
                            for tvar in e_spin_avg_var:
                                tvars.append(tvar)
                        if e_omni_spectra is not None:
                            tvars.append(e_omni_spectra)
                        if e_omni_spectra_spin is not None:
                            tvars.append(e_omni_spectra_spin)
                    elif datatype_id == 'extof':
                        # 9Feb2021, egrimes added 'helium' species for updates coming soon to the CDFs
                        p_spin_avg_var = mms_eis_spin_avg(probe=probe_id, species='proton', datatype=datatype_id, data_rate=data_rate_id, level=level_id, suffix=suffix)
                        o_spin_avg_var = mms_eis_spin_avg(probe=probe_id, species='oxygen', datatype=datatype_id, data_rate=data_rate_id, level=level_id, suffix=suffix)
                        a_spin_avg_var = mms_eis_spin_avg(probe=probe_id, species='alpha', datatype=datatype_id, data_rate=data_rate_id, level=level_id, suffix=suffix)
                        h_spin_avg_var = mms_eis_spin_avg(probe=probe_id, species='helium', datatype=datatype_id, data_rate=data_rate_id, level=level_id, suffix=suffix)
                        # create non-spin averaged omni-directional spectra
                        p_omni_spectra = mms_eis_omni(probe_id, species='proton', data_rate=data_rate_id, level=level_id, datatype=datatype_id, suffix=suffix)
                        o_omni_spectra = mms_eis_omni(probe_id, species='oxygen', data_rate=data_rate_id, level=level_id, datatype=datatype_id, suffix=suffix)
                        a_omni_spectra = mms_eis_omni(probe_id, species='alpha', data_rate=data_rate_id, level=level_id, datatype=datatype_id, suffix=suffix)
                        h_omni_spectra = mms_eis_omni(probe_id, species='helium', data_rate=data_rate_id, level=level_id, datatype=datatype_id, suffix=suffix)
                        # create spin averaged omni-directional spectra
                        p_omni_spectra_spin = mms_eis_omni(probe_id, species='proton', data_rate=data_rate_id, level=level_id, datatype=datatype_id, suffix=suffix+'_spin')
                        o_omni_spectra_spin = mms_eis_omni(probe_id, species='oxygen', data_rate=data_rate_id, level=level_id, datatype=datatype_id, suffix=suffix+'_spin')
                        a_omni_spectra_spin = mms_eis_omni(probe_id, species='alpha', data_rate=data_rate_id, level=level_id, datatype=datatype_id, suffix=suffix+'_spin')
                        h_omni_spectra_spin = mms_eis_omni(probe_id, species='helium', data_rate=data_rate_id, level=level_id, datatype=datatype_id, suffix=suffix+'_spin')
                        # add the vars to the output
                        if p_spin_avg_var is not None:
                            for tvar in p_spin_avg_var:
                                tvars.append(tvar)
                        if o_spin_avg_var is not None:
                            for tvar in o_spin_avg_var:
                                tvars.append(tvar)
                        if a_spin_avg_var is not None:
                            for tvar in a_spin_avg_var:
                                tvars.append(tvar)
                        if h_spin_avg_var is not None:
                            for tvar in h_spin_avg_var:
                                tvars.append(tvar)
                        if p_omni_spectra is not None:
                            tvars.append(p_omni_spectra)
                        if o_omni_spectra is not None:
                            tvars.append(o_omni_spectra)
                        if a_omni_spectra is not None:
                            tvars.append(a_omni_spectra)
                        if h_omni_spectra is not None:
                            tvars.append(h_omni_spectra)
                        if p_omni_spectra_spin is not None:
                            tvars.append(p_omni_spectra_spin)
                        if o_omni_spectra_spin is not None:
                            tvars.append(o_omni_spectra_spin)
                        if a_omni_spectra_spin is not None:
                            tvars.append(a_omni_spectra_spin)
                        if h_omni_spectra_spin is not None:
                            tvars.append(h_omni_spectra_spin)
                    elif datatype_id == 'phxtof':
                        # 9Feb2021, egrimes commented out oxygen calculations to match IDL updates
                        p_spin_avg_var = mms_eis_spin_avg(probe=probe_id, species='proton', datatype=datatype_id, data_rate=data_rate_id, level=level_id, suffix=suffix)
                        # o_spin_avg_var = mms_eis_spin_avg(probe=probe_id, species='oxygen', datatype=datatype_id, level=level_id, data_rate=data_rate_id, suffix=suffix)
                        # create non-spin averaged omni-directional spectra
                        p_omni_spectra = mms_eis_omni(probe_id, species='proton', data_rate=data_rate_id, datatype=datatype_id, level=level_id, suffix=suffix)
                        # o_omni_spectra = mms_eis_omni(probe_id, species='oxygen', data_rate=data_rate_id, level=level_id, datatype=datatype_id, suffix=suffix)
                        # create spin averaged omni-directional spectra
                        p_omni_spectra_spin = mms_eis_omni(probe_id, species='proton', data_rate=data_rate_id, datatype=datatype_id, level=level_id, suffix=suffix+'_spin')
                        # o_omni_spectra_spin = mms_eis_omni(probe_id, species='oxygen', data_rate=data_rate_id, level=level_id, datatype=datatype_id, suffix=suffix+'_spin')
                        # add the vars to the output
                        if p_spin_avg_var is not None:
                            for tvar in p_spin_avg_var:
                                tvars.append(tvar)
                        # if o_spin_avg_var is not None:
                        #     for tvar in o_spin_avg_var:
                        #         tvars.append(tvar)
                        if p_omni_spectra is not None:
                            tvars.append(p_omni_spectra)
                        # if o_omni_spectra is not None:
                        #     tvars.append(o_omni_spectra)
                        if p_omni_spectra_spin is not None:
                            tvars.append(p_omni_spectra_spin)
                        # if o_omni_spectra_spin is not None:
                        #     tvars.append(o_omni_spectra_spin)

                    mms_eis_set_metadata(tnames(tvars), data_rate=data_rate_id, datatype=datatype_id, suffix=suffix)

    return tnames(tvars)



import logging
import warnings
import numpy as np
from pytplot import get, store, options
from pyspedas.mms.feeps.mms_feeps_pitch_angles import mms_feeps_pitch_angles
from pyspedas.mms.feeps.mms_feeps_active_eyes import mms_feeps_active_eyes
from pyspedas.mms.feeps.mms_feeps_pad_spinavg import mms_feeps_pad_spinavg

# use nanmean from bottleneck if it's installed, otherwise use the numpy one
# bottleneck nanmean is ~2.5x faster
try:
    import bottleneck as bn
    nanmean = bn.nanmean
except ImportError:
    nanmean = np.nanmean

logging.captureWarnings(True)
logging.basicConfig(format='%(asctime)s: %(message)s', datefmt='%d-%b-%y %H:%M:%S', level=logging.INFO)


def mms_feeps_pad(bin_size=16.3636, probe='1', energy=[70, 600], level='l2', suffix='', datatype='electron', data_units='intensity', data_rate='srvy', angles_from_bfield=False):
    """
    This function will calculate pitch angle distributions using data from the MMS Fly's Eye Energetic Particle Sensor (FEEPS)
    
    Parameters
    ----------
        probe: str
            probe #, e.g., '4' for MMS4

        data_units: str
            'intensity' 

        datatype: str
            'electron' or 'ion'

        data_rate: str
            instrument data rate, e.g., 'srvy' or 'brst'

        level: str
            data level

        suffix: str
            suffix of the loaded data

        energy: list of float
            energy range to include in the calculation

        bin_size: float
            size of the pitch angle bins
            
        angles_from_bfield: bool
            calculate the pitch angles from the B-field data instead of reading from the CDFs

    Returns
    --------
        List of tplot variables created.
    """

    # account for angular response (finite field of view) of instruments
    # electrons can use +/- 21.4 deg on each pitch angle as average response angle; ions can start with +/-10 deg, but both need to be further refined
    if datatype == 'electron':
        dangresp = 21.4 # deg
    elif datatype == 'ion': 
        dangresp = 10.0 # deg

    if energy[0] < 32.0:
        logging.error('Please select a starting energy of 32 keV or above')
        return

    units_label = ''
    if data_units == 'intensity':
        units_label = '1/(cm^2-sr-s-keV)'
    elif data_units == 'counts':
        units_label = '[counts/s]'
    
    if not isinstance(probe, str):
        probe = str(probe)

    prefix = 'mms' + probe
    n_pabins = 180/bin_size
    pa_bins = [180.*pa_bin/n_pabins for pa_bin in range(0, int(n_pabins)+1)]
    pa_label = [180.*pa_bin/n_pabins+bin_size/2. for pa_bin in range(0, int(n_pabins))]

    if data_rate == 'brst' and angles_from_bfield == False:
        # v5.5+ = mms1_epd_feeps_srvy_l2_electron_pitch_angle
        pad_pas = get(prefix+'_epd_feeps_'+data_rate+'_'+level+'_'+datatype+'_pitch_angle'+suffix)
        if pad_pas is None:
            logging.error("Error reading variable containing FEEPS pitch angles")
            return
        pa_times = pad_pas[0]
        pa_data = pad_pas[1]
    else:
        pa_var, idx_maps = mms_feeps_pitch_angles(probe=probe, level=level, data_rate=data_rate, datatype=datatype, suffix=suffix)
        pa_times, pa_data = get(pa_var)

    if pa_data is None:
        logging.error("Error, couldn't find the PA variable")
        return

    eyes = mms_feeps_active_eyes([pa_times.min(), pa_times.max()], probe, data_rate, datatype, level)

    pa_data_map = {}

    if data_rate == 'srvy':
        if datatype == 'electron': 
            pa_data_map['top-electron'] = idx_maps['electron-top']
            pa_data_map['bottom-electron'] = idx_maps['electron-bottom']
        if datatype == 'ion':
            pa_data_map['top-ion'] = idx_maps['ion-top']
            pa_data_map['bottom-ion'] = idx_maps['ion-bottom']
    elif data_rate == 'brst':
        # note: the following are indices of the top/bottom sensors in pa_data
        # they should be consistent with pa_dlimits.labels
        pa_data_map['top-electron'] = [0, 1, 2, 3, 4, 5, 6, 7, 8]
        pa_data_map['bottom-electron'] = [9, 10, 11, 12, 13, 14, 15, 16, 17]
        # and ions:
        pa_data_map['top-ion'] = [0, 1, 2]
        pa_data_map['bottom-ion'] = [3, 4, 5]

    sensor_types = ['top', 'bottom']

    if datatype == 'electron':
        dflux = np.zeros([len(pa_times), len(pa_data_map['top-electron'])+len(pa_data_map['bottom-electron'])])
        dpa = np.zeros([len(pa_times), len(pa_data_map['top-electron'])+len(pa_data_map['bottom-electron'])])
    elif datatype == 'ion':
        dflux = np.zeros([len(pa_times), len(pa_data_map['top-ion'])+len(pa_data_map['bottom-ion'])])
        dpa = np.zeros([len(pa_times), len(pa_data_map['top-ion'])+len(pa_data_map['bottom-ion'])])

    for s_type in sensor_types:
        pa_map = pa_data_map[s_type+'-'+datatype]
        particle_idxs = [eye-1 for eye in eyes[s_type]]
        for isen, sensor_num in enumerate(particle_idxs):
            var_name = 'mms'+str(probe)+'_epd_feeps_'+data_rate+'_'+level+'_'+datatype+'_'+s_type+'_'+data_units+'_sensorid_'+str(sensor_num+1)+'_clean_sun_removed'+suffix
            times, data, energies = get(var_name)
            data[data == 0] = 'nan' # remove any 0s before averaging
            if np.isnan(energies[0]): # assumes all energies are NaNs if the first is
                continue
            # energy indices to use:
            indx = np.where((energies >= energy[0]) & (energies <= energy[1]))
            with warnings.catch_warnings():
                warnings.simplefilter("ignore", category=RuntimeWarning)
                dflux[:, pa_map[isen]] = nanmean(data[:, indx[0]], axis=1)
            dpa[:, pa_map[isen]] = pa_data[:, pa_map[isen]]

    # we need to replace the 0.0s left in after populating dpa with NaNs; these 
    # 0.0s are left in there because these points aren't covered by sensors loaded
    # for this datatype/data_rate
    dpa[dpa == 0] = 'nan'

    pa_flux = np.zeros([len(pa_times), int(n_pabins)])
    delta_pa = (pa_bins[1]-pa_bins[0])/2.0

    # Now loop through PA bins and time, find the telescopes where there is data in those bins and average it up!
    for pa_idx, pa_time in enumerate(pa_times):
        for ipa in range(0, int(n_pabins)):
            if not np.isnan(dpa[pa_idx, :][0]):
                with warnings.catch_warnings():
                    warnings.simplefilter("ignore", category=RuntimeWarning)
                    ind = np.where((dpa[pa_idx, :] + dangresp >= pa_label[ipa]-delta_pa) & (dpa[pa_idx, :]-dangresp < pa_label[ipa]+delta_pa))
                    if ind[0].size != 0:
                        if len(ind[0]) > 1:
                            pa_flux[pa_idx, ipa] = nanmean(dflux[pa_idx, ind[0]], axis=0)
                        else:
                            pa_flux[pa_idx, ipa] = dflux[pa_idx, ind[0]]

    pa_flux[pa_flux == 0] = 'nan' # fill any missed bins with NAN

    en_range_string = str(int(energy[0])) + '-' + str(int(energy[1])) + 'keV'
    new_name = 'mms'+probe+'_epd_feeps_'+data_rate+'_'+level+'_'+datatype+'_'+data_units+'_'+ en_range_string +'_pad'+suffix

    store(new_name, data={'x': times, 'y': pa_flux, 'v': pa_label})
    options(new_name, 'ylog', False)
    options(new_name, 'zlog', True)
    options(new_name, 'spec', True)
    options(new_name, 'ztitle', units_label)
    options(new_name, 'ytitle', 'MMS' + str(probe) + ' ' + datatype + ' PA')
    options(new_name, 'ysubtitle', '[deg]')

    # create the spin-averaged PAD
    spin_avg_var = mms_feeps_pad_spinavg(probe=probe, data_units=data_units, datatype=datatype, data_rate=data_rate, level=level, suffix=suffix, energy=energy, bin_size=bin_size)
    
    return [new_name, spin_avg_var]



import os
import unittest
from pytplot import data_exists
from pytplot import get_data
import pyspedas


class LoadTestCases(unittest.TestCase):
    def test_load_notplot(self):
        sem_vars = pyspedas.poes.sem(notplot=True)
        self.assertTrue('ted_ele_tel0_low_eflux' in sem_vars)

    def test_load_sem_data(self):
        sem_vars = pyspedas.poes.sem(time_clip=True)
        self.assertTrue(data_exists('ted_ele_tel0_low_eflux'))

    def test_load_probe_regression(self):
        sem_vars = pyspedas.poes.sem(probe='metop1', time_clip=True)
        self.assertTrue(data_exists('ted_ele_tel0_low_eflux'))
        m = get_data('ted_ele_tel0_low_eflux', metadata=True)
        self.assertTrue(m['CDF']['GATT']['Source_name'] == 'MetOp1')
        sem_vars = pyspedas.poes.sem(probe='metop2', time_clip=True)
        m = get_data('ted_ele_tel0_low_eflux', metadata=True)
        self.assertTrue(m['CDF']['GATT']['Source_name'] == 'MetOp2')

    def test_downloadonly(self):
        files = pyspedas.poes.sem(downloadonly=True, probe='noaa19')
        self.assertTrue(os.path.exists(files[0]))


if __name__ == '__main__':
    unittest.main()



from pyspedas.utilities.dailynames import dailynames
from pyspedas.utilities.download import download
from pytplot import time_clip as tclip
from pytplot import cdf_to_tplot

from .config import CONFIG

LANL_SC = {'l0': '90',
           'l1': '91',
           'l4': '94',
           'l7': '97',
           'l9': '89',
           'a1': '01a',
           'a2': '02a'}


def load(trange=['2004-10-31', '2004-11-01'], 
         instrument='mpa',
         probe='a1',
         level='k0',
         suffix='', 
         get_support_data=False, 
         varformat=None,
         varnames=[],
         downloadonly=False,
         notplot=False,
         no_update=False,
         time_clip=False):
    """
    This function loads data from the LANL mission; this function is not meant 
    to be called directly; instead, see the wrappers:

        pyspedas.lanl.mpa
        pyspedas.lanl.spa

    """

    probe = probe.lower()

    pathformat = LANL_SC[probe]+'_'+instrument+'/%Y/'+probe+'_'+level+'_'+instrument+'_%Y%m%d_v??.cdf'

    # find the full remote path names using the trange
    remote_names = dailynames(file_format=pathformat, trange=trange)

    out_files = []

    files = download(remote_file=remote_names, remote_path=CONFIG['remote_data_dir'], local_path=CONFIG['local_data_dir'], no_download=no_update)
    if files is not None:
        for file in files:
            out_files.append(file)

    out_files = sorted(out_files)

    if downloadonly:
        return out_files

    tvars = cdf_to_tplot(out_files, suffix=suffix, get_support_data=get_support_data, varformat=varformat, varnames=varnames, notplot=notplot)
    
    if notplot:
        return tvars

    if time_clip:
        for new_var in tvars:
            tclip(new_var, trange[0], trange[1], suffix='')

    return tvars

    


from .load import load
import numpy as np
import pandas as pd
from pytplot import store_data, options
from pyspedas import time_double
from pyspedas.cotrans.xyz_to_polar import xyz_to_polar


def pws(trange=['2012-10-01', '2012-10-02'],
        datatype='ne', 
        level='h1', 
        suffix='',  
        get_support_data=False, 
        varformat=None,
        varnames=[],
        downloadonly=False,
        notplot=False,
        no_update=False,
        time_clip=False):
    """
    This function loads data from the Plasma Waves and Sounder experiment (PWS)
    
    Parameters
    ----------
        trange : list of str
            time range of interest [starttime, endtime] with the format 
            'YYYY-MM-DD','YYYY-MM-DD'] or to specify more or less than a day 
            ['YYYY-MM-DD/hh:mm:ss','YYYY-MM-DD/hh:mm:ss']

        datatype: str
            Data type; Valid options:
                'ne', 'npw-ds', 'npw-py', 'spw'

        level: str
            Data level; options: 'h1' (default: h1)

        suffix: str
            The tplot variable names will be given this suffix.  By default, 
            no suffix is added.

        get_support_data: bool
            Data with an attribute "VAR_TYPE" with a value of "support_data"
            will be loaded into tplot.  By default, only loads in data with a 
            "VAR_TYPE" attribute of "data".

        varformat: str
            The file variable formats to load into tplot.  Wildcard character
            "*" is accepted.  By default, all variables are loaded in.

        varnames: list of str
            List of variable names to load (if not specified,
            all data variables are loaded)

        downloadonly: bool
            Set this flag to download the CDF files, but not load them into 
            tplot variables

        notplot: bool
            Return the data in hash tables instead of creating tplot variables

        no_update: bool
            If set, only load data from your local cache

        time_clip: bool
            Time clip the variables to exactly the range specified in the trange keyword

    Returns
    ----------
        List of tplot variables created.

    """
    tvars = load(instrument='pws', trange=trange, level=level, datatype=datatype, suffix=suffix, get_support_data=get_support_data, varformat=varformat, varnames=varnames, downloadonly=downloadonly, notplot=notplot, time_clip=time_clip, no_update=no_update)

    if tvars is None or notplot or downloadonly:
        return tvars

    return pws_postprocessing(tvars)


def pws_postprocessing(variables):
    """
    Placeholder for PWS post-processing 
    """
    return variables


def rdm(trange=['2012-10-01', '2012-10-02'],
        suffix='',  
        get_support_data=False, 
        varformat=None,
        varnames=[],
        downloadonly=False,
        notplot=False,
        no_update=False,
        time_clip=False):
    """
    This function loads data from the Radiation Moniter (RDM)
    
    Parameters
    ----------
        trange : list of str
            time range of interest [starttime, endtime] with the format 
            'YYYY-MM-DD','YYYY-MM-DD'] or to specify more or less than a day 
            ['YYYY-MM-DD/hh:mm:ss','YYYY-MM-DD/hh:mm:ss']

        suffix: str
            The tplot variable names will be given this suffix.  By default, 
            no suffix is added.

        get_support_data: bool
            Data with an attribute "VAR_TYPE" with a value of "support_data"
            will be loaded into tplot.  By default, only loads in data with a 
            "VAR_TYPE" attribute of "data".

        varformat: str
            The file variable formats to load into tplot.  Wildcard character
            "*" is accepted.  By default, all variables are loaded in.

        varnames: list of str
            List of variable names to load (if not specified,
            all data variables are loaded)

        downloadonly: bool
            Set this flag to download the CDF files, but not load them into 
            tplot variables

        notplot: bool
            Return the data in hash tables instead of creating tplot variables

        no_update: bool
            If set, only load data from your local cache

        time_clip: bool
            Time clip the variables to exactly the range specified in the trange keyword

    Returns
    ----------
        List of tplot variables created.

    """
    files = load(instrument='rdm', trange=trange, suffix=suffix, get_support_data=get_support_data, varformat=varformat, varnames=varnames, downloadonly=downloadonly, notplot=notplot, time_clip=time_clip, no_update=no_update)

    if files is None or notplot or downloadonly:
        return files

    return rdm_postprocessing(files)


def rdm_postprocessing(files):
    """
    Load the RDM ASCII files into tplot variables
    """
    data = load_csv_file(files)
    values = data.to_numpy()
    unix_times = time_double([ymd + '/' + hms for ymd, hms in zip(values[:, 0], values[:, 1])])

    L = np.float64(values[:, 2])
    INV = np.float64(values[:, 3])
    FMLAT = np.float64(values[:, 4])
    MLAT = np.float64(values[:, 5])
    MLT = np.float64(values[:, 6])
    ALT = np.float64(values[:, 7])
    GLAT = np.float64(values[:, 8])
    GLON = np.float64(values[:, 9])
    RDM_E3 = np.float64(values[:, 10])
    Energy = np.zeros(len(RDM_E3))
    Energy[:] = 2.5

    prefix_project = 'akb_'
    prefix_descriptor = 'rdm_'
    prefix = prefix_project + prefix_descriptor

    store_data(prefix_project+'L', data={'x': unix_times, 'y': L})
    store_data(prefix_project+'INV', data={'x': unix_times, 'y': INV})
    store_data(prefix_project+'FMLAT', data={'x': unix_times, 'y': FMLAT})
    store_data(prefix_project+'MLAT', data={'x': unix_times, 'y': MLAT})
    store_data(prefix_project+'MLT', data={'x': unix_times, 'y': MLT})
    store_data(prefix_project+'ALT', data={'x': unix_times, 'y': ALT})
    store_data(prefix_project+'GLAT', data={'x': unix_times, 'y': GLAT})
    store_data(prefix_project+'GLON', data={'x': unix_times, 'y': GLON})
    store_data(prefix+'FEIO', data={'x': unix_times, 'y': RDM_E3})
    store_data(prefix+'FEIO_Energy', data={'x': unix_times, 'y': Energy})

    options(prefix+'FEIO', 'spec', True)

    options(prefix_project+'L', 'ytitle', 'L-value')
    options(prefix_project+'INV', 'ytitle', 'Invariant Latitude [deg]')
    options(prefix_project+'FMLAT', 'ytitle', 'Footprint Latitude [deg]')
    options(prefix_project+'MLAT', 'ytitle', 'Magnetic Latitude [deg]')
    options(prefix_project+'MLT', 'ytitle', 'Magnetic Local Time [hour]')
    options(prefix_project+'ALT', 'ytitle', 'Altitude [km]')
    options(prefix_project+'GLAT', 'ytitle', 'Geographic Latitude [deg]')
    options(prefix_project+'GLON', 'ytitle', 'Geographic Longitude [deg]')
    options(prefix+'FEIO', 'ytitle', 'Omni-directional Integral Electron Flux')
    options(prefix+'FEIO', 'ysubtitle', '[/cm^22 sec str]')
    options(prefix+'FEIO_Energy', 'ytitle', 'Elctron energy [MeV]')

    return [prefix_project+'L',
            prefix_project+'INV',
            prefix_project+'FMLAT',
            prefix_project+'MLAT',
            prefix_project+'MLT',
            prefix_project+'ALT',
            prefix_project+'GLAT',
            prefix_project+'GLON',
            prefix+'FEIO',
            prefix+'FEIO_Energy']


def orb(trange=['2012-10-01', '2012-10-02'],
        suffix='',  
        get_support_data=False, 
        varformat=None,
        varnames=[],
        downloadonly=False,
        notplot=False,
        no_update=False,
        time_clip=False):
    """
    This function loads data from the Akebono orbit data (orb)
    
    Parameters
    ----------
        trange : list of str
            time range of interest [starttime, endtime] with the format 
            'YYYY-MM-DD','YYYY-MM-DD'] or to specify more or less than a day 
            ['YYYY-MM-DD/hh:mm:ss','YYYY-MM-DD/hh:mm:ss']

        suffix: str
            The tplot variable names will be given this suffix.  By default, 
            no suffix is added.

        get_support_data: bool
            Data with an attribute "VAR_TYPE" with a value of "support_data"
            will be loaded into tplot.  By default, only loads in data with a 
            "VAR_TYPE" attribute of "data".

        varformat: str
            The file variable formats to load into tplot.  Wildcard character
            "*" is accepted.  By default, all variables are loaded in.

        varnames: list of str
            List of variable names to load (if not specified,
            all data variables are loaded)

        downloadonly: bool
            Set this flag to download the CDF files, but not load them into 
            tplot variables

        notplot: bool
            Return the data in hash tables instead of creating tplot variables

        no_update: bool
            If set, only load data from your local cache

        time_clip: bool
            Time clip the variables to exactly the range specified in the trange keyword

    Returns
    ----------
        List of tplot variables created.

    """
    files = load(instrument='orb', trange=trange, suffix=suffix, get_support_data=get_support_data, varformat=varformat, varnames=varnames, downloadonly=downloadonly, notplot=notplot, time_clip=time_clip, no_update=no_update)

    if files is None or notplot or downloadonly:
        return files

    return orb_postprocessing(files)


def orb_postprocessing(files):
    """
    Load the orbit CSV files and create the tplot variables
    """
    prefix_project = 'akb_'
    prefix_descriptor = 'orb_'
    prefix = prefix_project + prefix_descriptor

    cols = ['pass','ut', 'ksc_azm', 'ksc_elv', 'ksc_dis', 'ksc_ang', 'syo_azm', 'syo_elv', 'syo_dis', 'syo_ang',
            'pra_azm', 'pra_elv', 'pra_dis', 'pra_ang', 'esr_azm', 'esr_elv', 'esr_dis', 'esr_ang', 'gclat','gclon',
            'inv', 'fmlat', 'mlat', 'mlt', 'bmdl_x', 'bmdl_y', 'bmdl_z', 'xxlon_sc', 'xxlat_sc', 'aheight','lsun',
            's_direc_x', 's_direc_y', 's_direc_z', 'sc_pos_x', 'sc_pos_y', 'sc_pos_z', 'sc_vel_x', 'sc_vel_y', 'sc_vel_z']

    data = load_csv_file(files, cols=cols)
    values = data.to_numpy()
    unix_times = time_double([date[2:4] + '-' + date[4:6] + '-' + date[0:2] + '/' + date[6:8] + ':' + date[8:10] + ':' + date[10:12] for date in data['ut']])

    km_in_re = 6374.4

    xyz = np.array([[data['sc_pos_x']], [data['sc_pos_y']], [data['sc_pos_z']]]).transpose([2, 0, 1]).squeeze()
    xyz = np.float64(xyz)
    xyz_re = xyz/km_in_re
    r_theta_phi = xyz_to_polar(xyz)
    rr = r_theta_phi[:, 0]
    th = r_theta_phi[:, 1]
    ph = r_theta_phi[:, 2]
    store_data(prefix + 'geo', data={'x': unix_times, 'y': xyz_re})
    store_data(prefix + 'gdlat', data={'x': unix_times, 'y': np.float64(data['gclat'])})
    store_data(prefix + 'gdlon', data={'x': unix_times, 'y': np.float64(data['gclon'])})
    store_data(prefix + 'inv', data={'x': unix_times, 'y': np.float64(data['inv'])})
    store_data(prefix + 'fmlat', data={'x': unix_times, 'y': np.float64(data['fmlat'])})
    store_data(prefix + 'MLT', data={'x': unix_times, 'y': np.float64(data['mlt'])})
    store_data(prefix + 'gcalt', data={'x': unix_times, 'y': rr / km_in_re})
    store_data(prefix + 'gclat', data={'x': unix_times, 'y': th})
    store_data(prefix + 'gclon', data={'x': unix_times, 'y': ph})
    options(prefix + 'geo', 'ytitle', 'GEO')
    options(prefix + 'geo', 'ysubtitle', '[Re]')
    options(prefix + 'gdlat', 'ytitle', 'Geodetic latitude of the magnetic footprint')
    options(prefix + 'gdlat', 'ysubtitle', '(120km altitude) [deg]')
    options(prefix + 'gdlon', 'ytitle', 'Geodetic longitude of the magnetic footprint')
    options(prefix + 'gdlon', 'ysubtitle', '(120km altitude) [deg]')
    options(prefix + 'inv', 'ytitle', 'Invariant Latitude of the magnetic footprint')
    options(prefix + 'inv', 'ysubtitle', '(120km altitude) [deg]')
    options(prefix + 'fmlat', 'ytitle', 'Geomagnetic Latitude of the magnetic footprint')
    options(prefix + 'fmlat', 'ysubtitle', '(120km altitude) [deg]')
    options(prefix + 'MLT', 'ytitle', 'Magnetic Local Time')
    options(prefix + 'MLT', 'ysubtitle', '[hours]')
    options(prefix + 'gcalt', 'ytitle', 'Geocentric Altitude')
    options(prefix + 'gcalt', 'ysubtitle', '[Re]')
    options(prefix + 'gclat', 'ytitle', 'Geocentric Latitude')
    options(prefix + 'gclat', 'ysubtitle', '[deg]')
    options(prefix + 'gclon', 'ytitle', 'Geocentric Longitude')
    options(prefix + 'gclon', 'ysubtitle', '[deg]')

    return [prefix + 'geo',
            prefix + 'gdlat',
            prefix + 'gdlon',
            prefix + 'inv',
            prefix + 'fmlat',
            prefix + 'MLT',
            prefix + 'gcalt',
            prefix + 'gclat',
            prefix + 'gclon']

        
def load_csv_file(filenames, cols=None):
    """
    Loads a list of CSV/txt files into pandas data frames
    """
    if not isinstance(filenames, list):
        filenames = [filenames]
    df = pd.concat((pd.read_csv(f, header=0, delim_whitespace=True, dtype=str, names=cols) for f in filenames), ignore_index=True)
    return df



from pytplot import get_data
from pyspedas.cotrans.cotrans_get_coord import cotrans_get_coord


def cotrans_set_coord(name, coord):
    """
    This function sets the coordinate system of a tplot variable, and updates legend names
    and axis labels if they include the coordinate system.

    Parameters:
        name: str
            name of the tplot variable

        coord: str
            Abbreviated name of the coordinate system (upper case recommended)

    Notes:
        The coordinate system is stored in the variable's metadata at:
            metadata['data_att']['coord_sys']

        See cotrans_get_coord to return the coordinate system

    Returns:
        bool: True/False depending on if the operation was successful
    """

    # check that the variable exists
    data = get_data(name)
    if data is None:
        return False

    metadata = get_data(name, metadata=True)

    # Get the current value of the coordinate system, if present
    coord_in = cotrans_get_coord(name)

    # note: updating the metadata dict directly updates
    # the variable's metadata in memory, so there's
    # no need to update the variable with store_data

    data_att = metadata.get('data_att')
    if data_att is None:
        metadata['data_att'] = {'coord_sys':coord}
    else:
        data_att['coord_sys'] = coord

    # should also update the legend, if it includes the coordinate system
    # for this to work, the coordinate system should be in all upper case
    if (coord_in is not None) and (metadata.get('plot_options') is not None):
        if metadata['plot_options'].get('yaxis_opt') is not None:
            if metadata['plot_options']['yaxis_opt'].get('legend_names') is not None:
                legend = metadata['plot_options']['yaxis_opt'].get('legend_names')
                updated_legend = [item.replace(coord_in.upper(), coord.upper()) for item in legend]
                metadata['plot_options']['yaxis_opt']['legend_names'] = updated_legend
            if metadata['plot_options']['yaxis_opt'].get('axis_label') is not None:
                ytitle = metadata['plot_options']['yaxis_opt'].get('axis_label')
                metadata['plot_options']['yaxis_opt']['axis_label'] = ytitle.replace(coord_in.upper(),
                                                                                     coord.upper())

    return True



from cdasws import CdasWs


def find_datasets(mission=None, instrument=None, label=False):
    cdas = CdasWs()
    datasets = cdas.get_datasets(observatoryGroup=mission)
    for index, dataset in enumerate(datasets):
        if instrument is not None:
            if instrument.upper() not in dataset['Id']:
                continue
        if label:
            print(dataset['Id'] + ': ' + dataset['Label'])
        else:
            print(dataset['Id'])



"""Test gmag and themis load functions."""
import os
import unittest
from pytplot import data_exists, time_string, time_double, del_data, get_timespan
from pyspedas.themis import autoload_support, get_spinmodel, fit
from pyspedas.themis.state.autoload_support import load_needed


class AutoLoadTestCases(unittest.TestCase):
    """Test themis support data autoload functions."""

    def test_load_needed_exact(self):
        # Test logic for determining whether support data needs to be loaded, based on exact comparisons
        # Timespan 1 completely encloses timespan 2
        ts1=['2007-03-23 00:00:00', '2007-03-24 00:00:00']
        ts2=['2007-03-23 00:00:01', '2007-03-23 23:59:59']

        tr1=time_double(ts1)
        tr2=time_double(ts2)
        trange_loaded=tr1
        trange_needed=tr2
        # trange loaded encloses trange needed on both sides, no load needed
        self.assertFalse(load_needed(trange_loaded,trange_needed,tolerance=0.0))
        trange_loaded=tr2
        trange_needed=tr1
        # trange loaded is a proper subset of trange needed, load needed
        self.assertTrue(load_needed(trange_loaded,trange_needed,tolerance=0.0))
        # tr3 starts before tr1
        tr3=[tr2[0] - 2.0, tr2[1]]
        trange_loaded=tr1
        trange_needed=tr3
        # trange needed starts before trange loaded, load needed
        self.assertTrue(load_needed(trange_loaded,trange_needed,tolerance=0.0))
        # tr4 ends after tr1
        tr4=[tr2[0], tr2[1]+2.0]
        trange_loaded=tr1
        trange_needed=tr4
        # trange needed ends after trange loaded, load needed
        self.assertTrue(load_needed(trange_loaded,trange_needed,tolerance=0.0))

    def test_load_needed_tolerance(self):
        # Test logic for determining whether support data needs to be loaded, while
        # allowing a certain amount of extrapolation
        tolerance = 10.0
        ts1=['2007-03-23 00:00:00', '2007-03-24 00:00:00']
        tr1 = time_double(ts1)
        trange_needed=tr1
        ts2=['2007-03-23 00:00:09', '2007-03-23 23:59:51']
        tr2=time_double(ts2)
        trange_loaded=tr2
        # Loaded data fails to overlap needed data by < tolerance on both sides, load not needed
        self.assertFalse(load_needed(trange_loaded,trange_needed,tolerance=tolerance))
        trange_loaded=tr1
        trange_needed=tr2
        # Loaded data overlaps needed data on both ends, load not needed
        self.assertFalse(load_needed(trange_loaded,trange_needed,tolerance=tolerance))
        trange_loaded=tr1
        trange_needed=[tr1[0] - 9.0, tr1[1]]
        # Loaded data fails to overlap needed data by < tolerance on left side, load not needed
        self.assertFalse(load_needed(trange_loaded,trange_needed,tolerance=tolerance))
        trange_needed=[tr1[0], tr1[1]+9.0]
        # Loaded data fails to overlap needed data by < tolerance on right side, load not needed
        self.assertFalse(load_needed(trange_loaded,trange_needed,tolerance=tolerance))
        trange_needed=[tr1[0] - 11.0, tr1[1]]
        # Loaded data fails to overlap needed data by > tolerance on left side, load needed
        self.assertTrue(load_needed(trange_loaded,trange_needed,tolerance=tolerance))
        trange_needed=[tr1[0], tr1[1]+11.0]
        # Loaded data fails to overlap needed data by > tolerance on right side, load needed
        self.assertTrue(load_needed(trange_loaded,trange_needed,tolerance=tolerance))


    def test_autoload_support_from_var(self):
        """Load FGM."""
        del_data('thc_*')
        del_data('slp_*')
        fit(trange=['2023-01-06','2023-01-07'])
        autoload_support(varname='thc_fgs_gse',slp=True,spinaxis=True,spinmodel=True)
        self.assertTrue(data_exists('thc_spinras'))
        self.assertTrue(data_exists('thc_spindec'))
        self.assertTrue(data_exists('slp_lun_att_x'))
        spinmodel=get_spinmodel(probe='c',correction_level=1)
        self.assertTrue(not (spinmodel is None))
        trange_needed=get_timespan('thc_fgs_gse')
        ts1=get_timespan('thc_spinras')
        ts2=get_timespan('thc_spindec')
        ts3=get_timespan('slp_lun_att_x')
        ts4=spinmodel.get_timerange()
        self.assertFalse(load_needed(ts1,trange_needed,tolerance=120.0))
        self.assertFalse(load_needed(ts2,trange_needed,tolerance=120.0))
        self.assertFalse(load_needed(ts3,trange_needed,tolerance=120.0))
        self.assertFalse(load_needed(ts4,trange_needed,tolerance=120.0))

    def test_autoload_support_without_var(self):
        """Load FGM."""
        del_data('thc_*')
        del_data('slp_*')
        trange=['2023-01-06','2023-01-07']
        autoload_support(trange=time_double(trange),probe='c',slp=True,spinaxis=True,spinmodel=True)
        self.assertTrue(data_exists('thc_spinras'))
        self.assertTrue(data_exists('thc_spindec'))
        self.assertTrue(data_exists('slp_lun_att_x'))
        spinmodel=get_spinmodel(probe='c',correction_level=1)
        self.assertTrue(not (spinmodel is None))
        trange_needed=time_double(trange)
        ts1=get_timespan('thc_spinras')
        ts2=get_timespan('thc_spindec')
        ts3=get_timespan('slp_lun_att_x')
        ts4=spinmodel.get_timerange()
        self.assertFalse(load_needed(ts1,trange_needed,tolerance=120.0))
        self.assertFalse(load_needed(ts2,trange_needed,tolerance=120.0))
        self.assertFalse(load_needed(ts3,trange_needed,tolerance=120.0))
        self.assertFalse(load_needed(ts4,trange_needed,tolerance=120.0))

    def test_autoload_support_reload_all(self):
        """Load FGM."""
        del_data('thc_*')
        del_data('slp_*')
        trange=['2008-01-01','2008-01-01']
        autoload_support(trange=time_double(trange),probe='c',slp=True,spinaxis=True,spinmodel=True)
        self.assertTrue(data_exists('thc_spinras'))
        self.assertTrue(data_exists('thc_spindec'))
        self.assertTrue(data_exists('slp_lun_att_x'))
        spinmodel=get_spinmodel(probe='c',correction_level=1)
        self.assertTrue(not (spinmodel is None))
        trange_needed=time_double(trange)
        ts1=get_timespan('thc_spinras')
        ts2=get_timespan('thc_spindec')
        ts3=get_timespan('slp_lun_att_x')
        ts4=spinmodel.get_timerange()
        self.assertFalse(load_needed(ts1,trange_needed,tolerance=120.0))
        self.assertFalse(load_needed(ts2,trange_needed,tolerance=120.0))
        self.assertFalse(load_needed(ts3,trange_needed,tolerance=120.0))
        self.assertFalse(load_needed(ts4,trange_needed,tolerance=120.0))
        # Now choose a different non-overlapping time range, and ensure everything got reloaded.
        trange=['2022-01-06','2022-01-07']
        autoload_support(trange=time_double(trange),probe='c',slp=True,spinaxis=True,spinmodel=True)
        trange_needed=time_double(trange)
        ts1=get_timespan('thc_spinras')
        ts2=get_timespan('thc_spindec')
        ts3=get_timespan('slp_lun_att_x')
        spinmodel=get_spinmodel(probe='c',correction_level=1)
        ts4=spinmodel.get_timerange()
        self.assertFalse(load_needed(ts1,trange_needed,tolerance=120.0))
        self.assertFalse(load_needed(ts2,trange_needed,tolerance=120.0))
        self.assertFalse(load_needed(ts3,trange_needed,tolerance=120.0))
        self.assertFalse(load_needed(ts4,trange_needed,tolerance=120.0))

    def test_autoload_support_var_doesntexist(self):
        # Should warn if a nonexistent tplot variable is passed
        autoload_support(varname='doesntexist',slp=True,spinaxis=True,spinmodel=True)

    def test_autoload_support_err_no_trange_no_var(self):
        # Should warn about trange being needed if no tplot variable passed
        autoload_support(probe='c',slp=True,spinaxis=True,spinmodel=True)

    def test_autoload_support_err_no_trange_no_probe(self):
        # Should warn about probe being needed if no tplot variable passed and spinaxis or spinmodel
        trange=['2007-03-23','2007-03-24']
        autoload_support(trange=trange,probe=None,slp=False,spinaxis=True,spinmodel=True)

    def test_autoload_support_no_probe_no_var_slp_only(self):
        # Passing a trange only should work if only SLP data is requested
        del_data('slp_*')
        trange=['2007-03-23','2007-03-24']
        autoload_support(trange=trange,slp=True)
        self.assertTrue(data_exists('slp_lun_att_x'))

if __name__ == '__main__':
    unittest.main()



"""Tests of ssl2dsl and dsl2gse functions."""

import unittest
from pytplot import get_data,del_data,tplot_restore,data_exists
from numpy.testing import assert_array_almost_equal_nulp, assert_array_max_ulp, assert_allclose
from pyspedas import cotrans_get_coord, cotrans_set_coord
from pyspedas.themis import autoload_support, ssl2dsl,dsl2gse


class DSLCotransDataValidation(unittest.TestCase):
    """ Compares cotrans results between Python and IDL """

    @classmethod
    def setUpClass(cls):
        """
        IDL Data has to be downloaded to perform these tests
        The SPEDAS script that creates the file: projects/themis/state/cotrans/thm_cotrans_validate.pro
        """
        from pyspedas.utilities.download import download
        from pyspedas.themis.config import CONFIG

        # Testing time range
        cls.t = ['2008-03-23', '2008-03-28']

        # Testing tolerance
        cls.tol = 1e-10

        # Download tplot files
        remote_server = 'https://spedas.org/'
        # remote_name = 'testfiles/thm_cotrans_validate.cdf'
        remote_name = 'testfiles/thm_cotrans_validate.tplot'
        datafile = download(remote_file=remote_name,
                            remote_path=remote_server,
                            local_path=CONFIG['local_data_dir'],
                            no_download=False)
        if not datafile:
            # Skip tests
            raise unittest.SkipTest("Cannot download data validation file")

        # Load validation variables from the test file
        del_data('*')
        filename = datafile[0]
        # pytplot.cdf_to_tplot(filename)
        tplot_restore(filename)
        #pytplot.tplot_names()
        cls.basis_x = get_data('basis_x')
        cls.basis_y = get_data('basis_y')
        cls.basis_z = get_data('basis_z')
        cls.basis_x_gei2gse = get_data('basis_x_gei2gse')
        cls.basis_y_gei2gse = get_data('basis_y_gei2gse')
        cls.basis_z_gei2gse = get_data('basis_z_gei2gse')
        cls.basis_x_gse2gei = get_data('basis_x_gse2gei')
        cls.basis_y_gse2gei = get_data('basis_y_gse2gei')
        cls.basis_z_gse2gei = get_data('basis_z_gse2gei')

        cls.basis_x_dsl2gse = get_data('basis_x_dsl2gse')
        cls.basis_y_dsl2gse = get_data('basis_y_dsl2gse')
        cls.basis_z_dsl2gse = get_data('basis_z_dsl2gse')

        cls.basis_x_gse2dsl = get_data('basis_x_gse2dsl')
        cls.basis_y_gse2dsl = get_data('basis_y_gse2dsl')
        cls.basis_z_gse2dsl = get_data('basis_z_gse2dsl')

        cls.basis_x_ssl2dsl = get_data('basis_x_ssl2dsl')
        cls.basis_y_ssl2dsl = get_data('basis_y_ssl2dsl')
        cls.basis_z_ssl2dsl = get_data('basis_z_ssl2dsl')

        cls.basis_x_dsl2ssl = get_data('basis_x_dsl2ssl')
        cls.basis_y_dsl2ssl = get_data('basis_y_dsl2ssl')
        cls.basis_z_dsl2ssl = get_data('basis_z_dsl2ssl')

        state_trange = ['2007-03-20', '2007-03-30']
        autoload_support(trange=state_trange, probe='a', spinaxis=True, spinmodel=True)


    def setUp(self):
        """ We need to clean tplot variables before each run"""
        # del_data('*')

    def test_gei2gse(self):
        """Validate gei2gse transform """
        from pyspedas.cotrans.cotrans_lib import subgei2gse

        bx = self.basis_x
        by = self.basis_y
        bz = self.basis_z
        times = bx.times
        bx_gse = subgei2gse(times, bx.y)
        by_gse = subgei2gse(times, by.y)
        bz_gse = subgei2gse(times, bz.y)
        assert_allclose(bx_gse, self.basis_x_gei2gse.y, atol=1.0e-06)
        assert_allclose(by_gse, self.basis_y_gei2gse.y, atol=1.0e-06)
        assert_allclose(bz_gse, self.basis_z_gei2gse.y, atol=1.0e-06)

    def test_gse2gei(self):
        """Validate gse2gei transform """
        from pyspedas.cotrans.cotrans_lib import subgse2gei

        bx = self.basis_x
        by = self.basis_y
        bz = self.basis_z
        times = bx.times
        bx_gei = subgse2gei(times, bx.y)
        by_gei = subgse2gei(times, by.y)
        bz_gei = subgse2gei(times, bz.y)
        assert_allclose(bx_gei, self.basis_x_gse2gei.y, atol=1.0e-06)
        assert_allclose(by_gei, self.basis_y_gse2gei.y, atol=1.0e-06)
        assert_allclose(bz_gei, self.basis_z_gse2gei.y, atol=1.0e-06)

    def test_dsl2gse_x(self):
        """Validate dsl2gse X axis transform """
        cotrans_set_coord('basis_x', 'DSL')
        result = dsl2gse('basis_x', 'tha_spinras_corrected', 'tha_spindec_corrected', 'basis_x_dsl2gse')
        self.assertEqual(result,1)
        bx_gse = get_data('basis_x_dsl2gse')
        assert_allclose(bx_gse.y, self.basis_x_dsl2gse.y, atol=1.0e-06)

    def test_dsl2gse_y(self):
        """Validate dsl2gse Y axis transform """
        from pyspedas.cotrans.cotrans_lib import subgse2gei
        cotrans_set_coord('basis_y', 'DSL')
        result = dsl2gse('basis_y', 'tha_spinras_corrected', 'tha_spindec_corrected', 'basis_y_dsl2gse')
        self.assertEqual(result, 1)
        by_gse = get_data('basis_y_dsl2gse')
        assert_allclose(by_gse.y, self.basis_y_dsl2gse.y, atol=1.0e-06)

    def test_dsl2gse_z(self):
        """Validate dsl2gse Z axis transform """
        cotrans_set_coord('basis_z', 'DSL')
        result = dsl2gse('basis_z', 'tha_spinras_corrected', 'tha_spindec_corrected', 'basis_z_dsl2gse')
        self.assertEqual(result, 1)
        bz_gse = get_data('basis_z_dsl2gse')
        assert_allclose(bz_gse.y, self.basis_z_dsl2gse.y, atol=1.0e-06)

    def test_gse2dsl_x(self):
        """Validate gse2dsl X axis transform """
        cotrans_set_coord('basis_x', 'GSE')
        result = dsl2gse('basis_x', 'tha_spinras_corrected', 'tha_spindec_corrected', 'basis_x_gse2dsl', isgsetodsl=True)
        self.assertEqual(result, 1)
        self.assertEqual(result, 1)
        bx_gse = get_data('basis_x_gse2dsl')
        assert_allclose(bx_gse.y, self.basis_x_gse2dsl.y, atol=1.0e-06)

    def test_gse2dsl_y(self):
        """Validate gse2dsl Y axis transform """
        from pyspedas.cotrans.cotrans_lib import subgse2gei
        cotrans_set_coord('basis_y', 'GSE')
        result = dsl2gse('basis_y', 'tha_spinras_corrected', 'tha_spindec_corrected', 'basis_y_gse2dsl', isgsetodsl=True)
        self.assertEqual(result, 1)
        by_gse = get_data('basis_y_gse2dsl')
        assert_allclose(by_gse.y, self.basis_y_gse2dsl.y, atol=1.0e-06)

    def test_gse2dsl_z(self):
        """Validate gse2dsl Z axis transform """
        cotrans_set_coord('basis_z', 'GSE')
        result = dsl2gse('basis_z', 'tha_spinras_corrected', 'tha_spindec_corrected', 'basis_z_gse2dsl', isgsetodsl=True)
        self.assertEqual(result, 1)
        bz_gse = get_data('basis_z_gse2dsl')
        assert_allclose(bz_gse.y, self.basis_z_gse2dsl.y, atol=1.0e-06)

    def test_ssl2dsl_x(self):
        """Validate ssl2dsl X axis transform """
        from pyspedas.themis.state.spinmodel.spinmodel import get_spinmodel
        sm = get_spinmodel(probe='a', correction_level=1)
        cotrans_set_coord('basis_x', 'SSL')
        result = ssl2dsl('basis_x', sm, 'basis_x_ssldsl')
        self.assertEqual(result, 1)
        bx_dsl = get_data('basis_x_ssl2dsl')
        assert_allclose(bx_dsl.y, self.basis_x_ssl2dsl.y, atol=1.0e-06)

    def test_ssl2dsl_y(self):
        """Validate ssl2dsl Y axis transform """
        from pyspedas.themis.state.spinmodel.spinmodel import get_spinmodel
        sm = get_spinmodel(probe='a', correction_level=1)
        cotrans_set_coord('basis_y', 'SSL')
        result = ssl2dsl('basis_y', sm, 'basis_y_ssldsl', use_spinphase_correction=True)
        self.assertEqual(result, 1)
        by_dsl = get_data('basis_y_ssl2dsl')
        assert_allclose(by_dsl.y, self.basis_y_ssl2dsl.y, atol=1.0e-06)

    def test_ssl2dsl_z(self):
        """Validate ssl2dsl Z axis transform """
        from pyspedas.themis.state.spinmodel.spinmodel import get_spinmodel
        sm = get_spinmodel(probe='a', correction_level=1)
        cotrans_set_coord('basis_z', 'SSL')
        result = ssl2dsl('basis_z', sm, 'basis_z_ssldsl', use_spinphase_correction=True)
        self.assertEqual(result, 1)
        bz_dsl = get_data('basis_z_ssl2dsl')
        assert_allclose(bz_dsl.y, self.basis_z_ssl2dsl.y, atol=1.0e-06)

    def test_dsl2ssl_x(self):
        """Validate dsl2ssl X axis transform """
        from pyspedas.themis.state.spinmodel.spinmodel import get_spinmodel
        sm = get_spinmodel(probe='a', correction_level=1)
        cotrans_set_coord('basis_x', 'DSL')
        result = ssl2dsl('basis_x', sm, 'basis_x_dsl2ssl', use_spinphase_correction=True, isdsltossl=True)
        self.assertEqual(result, 1)
        bx_ssl = get_data('basis_x_dsl2ssl')
        # This test needs a slightly looser tolerance for some reason.
        assert_allclose(bx_ssl.y, self.basis_x_dsl2ssl.y, atol=1.5e-06)

    def test_dsl2ssl_y(self):
        """Validate dsl2ssl Y axis transform """
        from pyspedas.themis.state.spinmodel.spinmodel import get_spinmodel
        sm = get_spinmodel(probe='a', correction_level=1)
        cotrans_set_coord('basis_y', 'DSL')
        result = ssl2dsl('basis_y', sm, 'basis_y_dsl2ssl', use_spinphase_correction=True, isdsltossl=True)
        self.assertEqual(result, 1)
        by_ssl = get_data('basis_y_dsl2ssl')
        # This test needs a slightly looser tolerance for some reason.
        assert_allclose(by_ssl.y, self.basis_y_dsl2ssl.y, atol=1.5e-06)

    def test_dsl2ssl_z(self):
        """Validate dsl2ssl Z axis transform """
        from pyspedas.themis.state.spinmodel.spinmodel import get_spinmodel
        sm = get_spinmodel(probe='a', correction_level=1)
        cotrans_set_coord('basis_z', 'DSL')
        result = ssl2dsl('basis_z', sm, 'basis_z_dsl2ssl', use_spinphase_correction=True, isdsltossl=True)
        self.assertEqual(result, 1)
        bz_ssl = get_data('basis_z_dsl2ssl')
        assert_allclose(bz_ssl.y, self.basis_z_dsl2ssl.y, atol=1.0e-06)

    def test_catch_mismatch_dsl2ssl_z(self):
        """Test detection of mismatched input vs. requested coordinate systems in dsl2ssl transform """
        from pyspedas.themis.state.spinmodel.spinmodel import get_spinmodel
        sm = get_spinmodel(probe='a', correction_level=1)
        # Requesting DSL to SSL, but specifying SSL as input coordinate system
        cotrans_set_coord('basis_z', 'SSL')
        result = ssl2dsl('basis_z', sm, 'basis_z_dsl2ssl', use_spinphase_correction=True, isdsltossl=True, ignore_input_coord = False)
        self.assertEqual(result, 0)

    def test_catch_mismatch_ssl2dsl_z(self):
        """Test detection of mismatched input vs. requested coordinates in ssl2dsl transform """
        from pyspedas.themis.state.spinmodel.spinmodel import get_spinmodel
        sm = get_spinmodel(probe='a', correction_level=1)
        # Requesting SSL to DSL, but specifying DSL as input coordinate system
        cotrans_set_coord('basis_z', 'DSL')
        result = ssl2dsl('basis_z', sm, 'basis_z_ssldsl', use_spinphase_correction=True, ignore_input_coord=False)
        self.assertEqual(result, 0)

    def test_catch_mismatch_gse2dsl_z(self):
        """Test detection of mismatched input vs requested coordinates in gse2dsl transform """
        cotrans_set_coord('basis_z', 'DSL')
        result = dsl2gse('basis_z', 'tha_spinras_corrected', 'tha_spindec_corrected', 'basis_z_gse2dsl', isgsetodsl=True)
        self.assertEqual(result, 0)

    def test_catch_mismatch_dsl2gse_z(self):
        """Test detection of mismatched input vs. requested coordinates in dsl2gse transform """
        cotrans_set_coord('basis_z', 'GSE')
        result = dsl2gse('basis_z', 'tha_spinras_corrected', 'tha_spindec_corrected', 'basis_z_gse2dsl')
        self.assertEqual(result, 0)

    def test_ignore_mismatch_dsl2ssl_z(self):
        """Test ability to bypass coordinate system consistency check in dsl2ssl transform """
        from pyspedas.themis.state.spinmodel.spinmodel import get_spinmodel
        sm = get_spinmodel(probe='a', correction_level=1)
        # Requesting DSL to SSL, but specifying SSL as input coordinate system
        cotrans_set_coord('basis_z', 'SSL')
        result = ssl2dsl('basis_z', sm, 'basis_z_dsl2ssl', use_spinphase_correction=True, isdsltossl=True, ignore_input_coord = True)
        self.assertEqual(result, 1)

    def test_ignore_mismatch_ssl2dsl_z(self):
        """Test ability to bypass coordinate system consistency check in ssl2dsl transform  """
        from pyspedas.themis.state.spinmodel.spinmodel import get_spinmodel
        sm = get_spinmodel(probe='a', correction_level=1)
        # Requesting SSL to DSL, but specifying DSL as input coordinate system
        cotrans_set_coord('basis_z', 'DSL')
        result = ssl2dsl('basis_z', sm, 'basis_z_ssldsl', use_spinphase_correction=True, ignore_input_coord=True)
        self.assertEqual(result, 1)

    def test_ignore_mismatch_gse2dsl_z(self):
        """Test ability to bypass coordinate system consistency check in gse2dsl transform """
        cotrans_set_coord('basis_z', 'DSL')
        result = dsl2gse('basis_z', 'tha_spinras_corrected', 'tha_spindec_corrected', 'basis_z_gse2dsl', isgsetodsl=True, ignore_input_coord=True)
        self.assertEqual(result, 1)

    def test_ignore_mismatch_dsl2gse_z(self):
        """Test ability to bypass coordinate system consistency check in dsl2gse transform """
        cotrans_set_coord('basis_z', 'GSE')
        result = dsl2gse('basis_z', 'tha_spinras_corrected', 'tha_spindec_corrected', 'basis_z_gse2dsl', ignore_input_coord=True)
        self.assertEqual(result, 1)

if __name__ == '__main__':
    unittest.main()



import cdflib
import logging
import numpy as np


from copy import deepcopy
from scipy.spatial import KDTree
from pyspedas import tnames
from pytplot import time_double
from pytplot import time_string
from pytplot import get_data
from scipy import interpolate

from astropy.coordinates import spherical_to_cartesian, cartesian_to_spherical

logging.captureWarnings(True)
logging.basicConfig(format='%(asctime)s: %(message)s',
                    datefmt='%d-%b-%y %H:%M:%S', level=logging.INFO)
def erg_lepe_get_dist(tname,
                      index=None,
                      units='flux',
                      level='l2',
                      species='e',
                      time_only=False,
                      single_time=None,
                      trange=None):
    if len(tnames(tname)) > 0:
        input_name = tnames(tname)[0]
    else:
        print(f'Variable: {tname} not found!')
        return 0
    level = level.lower()
    """
    ;; Extract some information from a tplot variable name
    ;; e.g., erg_lepi_l2_3dflux_FPDU
    """
    vn_info = input_name.split('_')
    instrument = vn_info[1]
    level = vn_info[2]
    arrnm = vn_info[4]
    if species is None:
        if instrument == 'lepe':
            species = 'e'
        else:
            print(f'RROR: given an invalid tplot variable: {input_name}')
    # ;; Get a reference to data and metadata
    data_in = get_data(input_name)
    data_in_metadata = get_data(input_name, metadata=True)
    if data_in is None:
        print('Problem extracting the mepe 3dflux data.')
        return 0
    # ;; Return time labels
    if time_only:
        return data_in[0]
    if single_time is not None:
        single_time_double = time_double(single_time)
        if (data_in[0][0] <= single_time_double)\
                and (single_time_double <= data_in[0][-1]):
            nearest_time = interpolate.interp1d(data_in[0], data_in[0],
                                                kind="nearest")(single_time_double)
        elif (single_time_double < data_in[0][0])\
                or (data_in[0][-1] < single_time_double):
            nearest_time = interpolate.interp1d(data_in[0], data_in[0],
                                                kind="nearest", fill_value='extrapolate')(single_time_double)
        index = np.where(data_in[0] == nearest_time)[0]
        n_times = index.size
    else:
        # index supersedes time range
        if index is None:
            if trange is not None:
                trange_double_array = np.array(time_double(trange))
                trange_minmax = np.array([trange_double_array.min(),
                                          trange_double_array.max()])
                index = np.where((data_in[0] >= trange_minmax[0])
                                 & (data_in[0] <= trange_minmax[1]))[0]
                n_times = index.size
                if n_times == 0:
                    print('No data in time range: '
                          + ' '.join(time_string(trange_minmax)))
                    return 0
            else:
                n_times = data_in[0].size
                index = np.arange(n_times)
        else:
            n_times = np.array([index]).size
    """
    ;; --------------------------------------------------------------
    ;; LEP-e data Array[10269(time), 32(energy), 12(anode), 16(spin
    ;; phase)] for the normal 3-D flux mode (fine channels are averaged
    ;;into two coarse channels)
    """
    # ;; to [ energy, spin phase(azimuth), apd(elevation) ]
    dim_array = np.array(data_in[1].shape[1:])[[0, 2, 1]]
    if species.lower() == 'e':
        mass = 5.68566e-06
        charge = -1.
        data_name = 'LEP-e Electron 3dflux'
        integ_time = 7.99 / 32 / 16  # ;; currently hard-coded
    else:
        print('given species is not supported by this routine.')
        return 0
    #  ;; basic template structure compatible with other routines
    dist = {
        'project_name': 'ERG',
        'spacecraft': 1,  # always 1 as a dummy value
        'data_name': data_name,
        'units_name': 'flux',  # ; LEP-e data in [/eV-s-sr-cm2
        'units_procedure': 'erg_convert_flux_units',
        'species': species,
        'valid': 1,
        'charge': charge,
        'mass': mass,
    }
    # other keys (members) will added in below processing, sequentially.
    """
    ;; Then, fill in arrays in the data structure
    ;;   dim[ nenergy, nspinph(azimuth), nanode(elevation), ntime]
    """
    dist['time'] = data_in[0][[index]]
    dist['end_time'] = dist['time'] + integ_time  # ;; currently hard-coded
    """
    ;; Shuffle the original data array [time,energy, anode, spin phase] to
    ;; be energy-azimuth(spin phase)-elevation-time.
    """
    dist['data'] = data_in[1][[index]].transpose([1, 3, 2, 0])
    dist['bins'] = np.ones(shape=np.insert(dim_array, dim_array.shape[0],
                                           n_times), dtype='int8')
    # must be set or data will be consider invalid
    # ;; fishy negative flux values are all replaced with zero.
    # ;; dist['data'] = np.where(dist['data'] < 0.,0.,dist['data'])
    file_name_raw= get_data(input_name, metadata=True)['CDF']['FILENAME']
    if isinstance(file_name_raw, str):
        cdf_path = file_name_raw
    elif isinstance(file_name_raw, list):
        cdf_path = file_name_raw[-1]

    cdf_file = cdflib.CDF(cdf_path)

    #  ;; Energy ch
    """
    ;; Default unit of v in F?DU tplot variables [keV/q] should be
    ;; converted to [eV] by multiplying (1000 * charge number).
    """
    e0_array_raw = data_in[2][[index]]# ;; [time, 32]
    e0_array = e0_array_raw.T
    energy_reform = np.reshape(e0_array, [dim_array[0], 1, 1, n_times])
    energy_rebin1 = np.repeat(energy_reform, dim_array[2],
                             axis=2)  # repeated across apd(elevation)
    dist['energy'] = np.repeat(energy_rebin1, dim_array[1],
                              axis=1)  # repeated across spin phase(azimuth)

    # denergy member
    
    de_array = e0_array + np.nan  #  ;; initialized as [ time, 32 ]  
    for i in range(n_times):
        enec0_array = deepcopy(e0_array[:, i])
        id_array = np.argwhere(np.isfinite(enec0_array))
        if len(id_array) < 2:
            continue
        #  ;sorting and picks up only uniq elements. nominally 30, 28, 5, 4.
        enec_array = np.sort(np.unique(enec0_array[id_array[:,0].tolist()]))
        n_enec = enec_array.size
        if (n_enec < 3) or (np.nansum(enec_array) < 0.):
            continue  # ;; invalid energy value found
        logenec_array = np.log10(enec_array)
        logmn_array = (logenec_array[1:] + logenec_array[:-1]) / 2.
        logdep_array = enec_array * 0.
        logdem_array = enec_array * 0.
        logdep_array[:-1] = deepcopy(logmn_array)
        logdem_array[1:] = deepcopy(logmn_array)
        logdem_array[0] = logenec_array[0] - (logmn_array[0] - logenec_array[0])
        logdep_array[-1] = logenec_array[-1] + (logenec_array[-1] - logmn_array[-1])
        de_array_i= 10.**logdep_array - 10.**logdem_array

        # getting nearest neighbor indices, 'id = nn( enec, enec0 ) ' in IDL
        enec0_array_temp = np.nan_to_num(enec0_array, copy=True,nan=np.nanmin(enec0_array))
        enec0_array_temp_repeat= np.repeat(np.array([enec0_array_temp]).T, 2, 1)
        enec_array_repeat= np.repeat(np.array([enec_array]).T, 2, 1)
        tree = KDTree(enec_array_repeat, leafsize=1)
        _, id_array = tree.query(enec0_array_temp_repeat)
        # getting nearest neighbor indices, 'id = nn( enec, enec0 ) ' in IDL
        de_array[:, i] = de_array_i[id_array].T

    de_reform = np.reshape(de_array, [dim_array[0], 1, 1, n_times])
    de_rebin1 = np.repeat(de_reform, dim_array[2],
                         axis=2)  # repeated across apd(elevation)
    dist['denergy'] = np.repeat(de_rebin1, dim_array[1],
                         axis=1)  # repeated across spin phase(azimuth)

    dist['n_energy'] = dim_array[0]
    dist['n_bins'] = dim_array[1] * dim_array[2]  #   # thetas * # phis
    """
    ;; Array elements containing NaN are excluded from the further
    ;; calculations, by setting bins to be zero.
    """
    dist['bins'] = np.where((np.isfinite(dist['data']))
                               &(np.isfinite(dist['energy'])),
                               1,0)
    
    #  ;; azimuthal angle in spin direction
    angarr = cdf_file.varget('FEDU_Angle_SGA') # ;;[elev/phi, (anode)] in SGA  (looking dir)

    # ;; Flip the looking dirs to the flux dirs
    n_anode = angarr[0].size
    r_array = np.ones(n_anode)
    elev_array = deepcopy(angarr[0])
    phi_array = deepcopy(angarr[1])
    deg_to_rad = np.pi / 180.
    x_array, y_array, z_array = spherical_to_cartesian(
                                     r_array,
                                     elev_array * deg_to_rad,
                                     phi_array* deg_to_rad
                                     )
    r_array, elev_array, phi_array = cartesian_to_spherical(
                                        -x_array,
                                        -y_array,
                                        -z_array)
    elev_array = elev_array.value / deg_to_rad
    phi_array = phi_array.value / deg_to_rad
    angarr = np.array([elev_array, phi_array])

    phissi = angarr[1] - (90. + 21.6)  # ;; [(anode)] (21.6 = degree between sun senser and satellite coordinate)
    spinph_ofst = data_in[4] * 22.5
    phi0_1_reform = np.reshape(phissi, [1, 1, dim_array[2], 1])
    phi0_1_rebin1 = np.repeat(phi0_1_reform, dim_array[1],
                             axis=1)  # repeated across spin phase(azimuth)
    phi0_1_rebin2 = np.repeat(phi0_1_rebin1, dim_array[0],
                             axis=0)  # repeated across energy
    phi0_1 = np.repeat(phi0_1_rebin2,     n_times,
                              axis=3)  # repeated across n_times
    phi0_2_reform = np.reshape(spinph_ofst, [1, dim_array[1], 1, 1])
    phi0_2_rebin1 = np.repeat(phi0_2_reform, dim_array[2],
                             axis=2)  # repeated across apd(elevation)
    phi0_2_rebin2 = np.repeat(phi0_2_rebin1, dim_array[0],
                             axis=0)  # repeated across energy
    phi0_2 = np.repeat(phi0_2_rebin2, n_times,
                              axis=3)  # repeated across n_times
    phi0 = phi0_1 + phi0_2
    ofst_sv = (np.arange(dim_array[0]) + 0.5) * \
        22.5/dim_array[0]  # ;; [(energy)]
    phi_ofst_for_sv_reform = np.reshape(ofst_sv, [dim_array[0], 1, 1, 1])
    phi_ofst_for_sv_rebin1 = np.repeat(
        phi_ofst_for_sv_reform, dim_array[2],
                                 axis=2)  # repeated across apd(elevation)
    phi_ofst_for_sv_rebin2 = np.repeat(
        phi_ofst_for_sv_rebin1,dim_array[1],
                                 axis=1)  # repeated across spin phase(azimuth)
    phi_ofst_for_sv = np.repeat(phi_ofst_for_sv_rebin2,         n_times,
                                 axis=3)  # repeated across n_times
    """
    ;;  phi angle for the start of each spin phase
    ;;    + offset angle foreach sv step
    """
    dist['phi'] = np.fmod((phi0 + phi_ofst_for_sv + 360.), 360.)
    dist['dphi'] = np.full(shape=np.insert(dim_array, dim_array.shape[0],
                                           n_times), fill_value=22.5)  # ;; 22.5 deg as a constant
    del phi0, phi_ofst_for_sv  # ;; Clean huge arrays
    dist['n_phi'] = dim_array[1]
    #  ;; elevation angle
    elev = angarr[0]  # ;; [(anode)]
    elev_reform = np.reshape(elev, [1, 1, dim_array[2], 1])
    elev_rebin1 = np.repeat(elev_reform, dim_array[1],
                             axis=1)  # repeated across spin phase(azimuth)
    elev_rebin2 = np.repeat(elev_rebin1, dim_array[0],
                             axis=0)  # repeated across energy
    dist['theta'] = np.repeat(elev_rebin2, n_times,
                             axis=3)  # repeated across n_times

    #  ;; elevation angle for fine channel  
    
    if 'fine_ch' in vn_info:
        dist['dtheta'] = np.full(shape=np.insert(dim_array, dim_array.shape[0],
                                             n_times), fill_value=22.5/6.)  #  ;; Fill all with 22.5/6.
    else:
        dist['dtheta'] = np.full(shape=np.insert(dim_array, dim_array.shape[0],
                                             n_times), fill_value=22.5)  #  ;; Fill all with 22.5. first
        #  ;; give half weight for ch1,2,3,4, 19,20,21,22
        dist['dtheta'][:, :, 0:4, :] = 11.25
        dist['dtheta'][:, :, dim_array[2]-4:dim_array[2], :] = 11.25
        if dim_array[2] == 22:  #  ;; with full fine channels
            dist['dtheta'][:, :, 5:17, :] = 22.5/6

    dist['n_theta'] = dim_array[2]
    
    return dist



import unittest
from pytplot import data_exists
import pyspedas


class LoadTestCases(unittest.TestCase):
    def test_load_mfi_data(self):
        mfi_vars = pyspedas.ace.mfi(trange=['2018-11-5', '2018-11-6'], time_clip=True)
        self.assertTrue(data_exists('Magnitude'))

    def test_load_mfi_notplot(self):
        mfi_vars = pyspedas.ace.mfi(trange=['2018-11-5', '2018-11-6'], notplot=True)
        self.assertTrue(isinstance(mfi_vars, dict))

    def test_load_mfi_downloadonly(self):
        mfi_vars = pyspedas.ace.mfi(trange=['2018-11-5', '2018-11-6'], downloadonly=True)
        self.assertTrue(isinstance(mfi_vars, list))

    def test_load_swe_data(self):
        swe_vars = pyspedas.ace.swe()
        self.assertTrue(data_exists('Np'))
        self.assertTrue(data_exists('Vp'))

    def test_load_epm_data(self):
        epm_vars = pyspedas.ace.epam()
        self.assertTrue(data_exists('H_lo'))
        self.assertTrue(data_exists('Ion_very_lo'))
        self.assertTrue(data_exists('Ion_mid'))
        self.assertTrue(data_exists('Electron_hi'))

    def test_load_cris_data(self):
        cris_vars = pyspedas.ace.cris()
        self.assertTrue(data_exists('flux_N'))

    def test_load_sis_data(self):
        sis_vars = pyspedas.ace.sis()
        self.assertTrue(data_exists('H_lo'))
        self.assertTrue(data_exists('H_hi'))

    def test_load_ule_data(self):
        ule_vars = pyspedas.ace.uleis()
        self.assertTrue(data_exists('H_S1'))

    def test_load_sep_data(self):
        sep_vars = pyspedas.ace.sepica()
        self.assertTrue(data_exists('H1'))
        self.assertTrue(data_exists('H2'))
        self.assertTrue(data_exists('H3'))

    def test_load_swi_data(self):
        swi_vars = pyspedas.ace.swics()
        self.assertTrue(data_exists('vHe2'))

    def test_data_dir(self):
        self.assertTrue(pyspedas.ace.config.CONFIG['local_data_dir'] == 'ace_data/')


if __name__ == '__main__':
    unittest.main()




import logging

import numpy as np
from copy import deepcopy
from pyspedas import tnames
from pytplot import time_double
from pytplot import time_string
from pytplot import get_data
from scipy import interpolate


logging.captureWarnings(True)
logging.basicConfig(format='%(asctime)s: %(message)s',
                    datefmt='%d-%b-%y %H:%M:%S', level=logging.INFO)


def erg_xep_get_dist(tname,
                      index=None,
                      units='flux',
                      level='l2',
                      species='e',
                      time_only=False,
                      single_time=None,
                      trange=None):

    if len(tnames(tname)) > 0:
        input_name = tnames(tname)[0]
    else:
        print(f'Variable: {tname} not found!')
        return 0

    level = level.lower()
    """
    ;; Extract some information from a tplot variable name
    ;; e.g., erg_xep_l2_FEDU_SSD
    """

    vn_info = input_name.split('_')
    instrument = vn_info[1]
    level = vn_info[2]

    if instrument == 'xep':
        species = 'e'
    else:
        print(f'ERROR: given an invalid tplot variable: {input_name}')
        return 0

    # ;; Get a reference to data and metadata

    data_in = get_data(input_name)
    data_in_metadata = get_data(input_name, metadata=True)

    if data_in is None:
        print('Problem extracting the mepe 3dflux data.')
        return 0
    else:
        if len(data_in) != 4:
            print(f'Variable: {input_name} contains wrong number of elements!')

    # ;; Return time labels
    if time_only:
        return data_in[0]

    #  ;; Estimate the spin periods
    t_phtime = deepcopy(data_in[0])
    sc0_dt = t_phtime[1:] - t_phtime[:-1]
    sc0_dt = np.insert(sc0_dt, sc0_dt.shape[0], sc0_dt[-1])
    n_times = len(t_phtime)
    sctintgt = np.repeat(np.array([sc0_dt / 16.]).T, 16, axis=1)
    phtime = np.repeat(np.array([t_phtime]).T, 16, axis=1)\
            + sctintgt\
            * np.repeat(np.array([np.arange(16,dtype = 'float')]),
                        n_times, axis=0)


    if single_time is not None:
        single_time_double = time_double(single_time)
        if (data_in[0][0] <= single_time_double)\
                and (single_time_double <= data_in[0][-1]):
            nearest_time = interpolate.interp1d(data_in[0], data_in[0],
                                                kind="nearest")(single_time_double)
        elif (single_time_double < data_in[0][0])\
                or (data_in[0][-1] < single_time_double):
            nearest_time = interpolate.interp1d(data_in[0], data_in[0],
                                                kind="nearest", fill_value='extrapolate')(single_time_double)
        index = np.where(data_in[0] == nearest_time)[0]
        n_times = index.size
    else:
        # index supersedes time range
        if index is None:
            if trange is not None:
                trange_double_array = np.array(time_double(trange))
                trange_minmax = np.array([trange_double_array.min(),
                                          trange_double_array.max()])
                index = np.where((data_in[0] >= trange_minmax[0])
                                 & (data_in[0] <= trange_minmax[1]))[0]
                n_times = index.size
                if n_times == 0:
                    print('No data in time range: '
                          + ' '.join(time_string(trange_minmax)))
                    return 0
            else:
                n_times = data_in[0].size
                index = np.arange(n_times)
        else:
            n_times = np.array([index]).size

    """
    ;; --------------------------------------------------------------

    ;; XEP data arr: [10080(time), 9(energy), 16(spin ph)]
    ;; Dimensions
    """

    dim_array = np.array(data_in[1].shape[1:])  #  ;; [energy, spin phase(azimuth) ]
    # ;;dim_array[0] = 12  #  ;; Use only the SSD channels for now

    if species.lower() == 'e':
        mass = 5.68566e-06
        charge = -1.
        data_name = 'XEP Electron 2dflux'
        integ_time = 7.99 / 16  # ;; currently hard-coded
    else:
        print('given species is not supported by this routine.')
        return 0

    #  ;; basic template structure compatible with other routines

    dist = {
        'project_name': 'ERG',
        'spacecraft': 1,  # always 1 as a dummy value
        'data_name': data_name,
        'units_name': 'flux',  # MEP-e data in [/keV-s-sr-cm2]
                               # should be converted to [/eV-s-sr-cm2]
        'units_procedure': 'erg_convert_flux_units',
        'species': species,
        'valid': 1,

        'charge': charge,
        'mass': mass,

    }

    # other keys (members) will added in below processing, sequentially.

    """
    ;; Then, fill in arrays in the data structure
    ;;   dim[ nenergy, nspinph(azimuth), napd(elevation), ntime]
    """

    dist['time'] = data_in[0][tuple([index])]
    dist['end_time'] = dist['time'] + integ_time  # ;; currently hard-coded

    """
    ;; Shuffle the original data array [time,energy,spin phase] to
    ;; be energy-azimuth-time.
    ;; The factor 1d-3 is to convert [/keV-s-sr-cm2] (default unit of
    ;; XEP Lv2 2dflux data) to [/eV-s-sr-cm2].
    ;; Again only SSD channels are used (0:11 for ene. ch.).
    """
    dist['data'] = data_in[1][tuple([index])].reshape(
                    n_times, dim_array[0], dim_array[1]).transpose([1, 2, 0]) * 1e-3

    dist['bins'] = np.ones(shape=np.insert(dim_array, dim_array.shape[0],
                                           n_times), dtype='int8')
    # must be set or data will be consider invalid

    """
    ;; No invalid ch is set for XEP currently.
    """
    #  ;; dist['bins'][0] = 0

    #  ;; Energy ch
    e0_array = data_in[2] * 1e+3  # ;; [MeV] 
                            # (default of XEP Lv2 FEDU tplot var) to [eV]
    energy_reform = np.reshape(e0_array, [dim_array[0], 1, 1])
    energy_rebin1 = np.repeat(energy_reform, dim_array[1],
                             axis=1)  # repeated across apd(elevation)
    dist['energy'] =  np.repeat(energy_rebin1, n_times,
                              axis=2)  # repeated across n_times

    #  ;; Energy bin width
    e0bnd = np.sqrt(e0_array[:-1] * e0_array[1:])  # ;;[8]
    # ;; [9]  upper boundary of energy bin
    e0bnd_p = np.insert(e0bnd, e0bnd.shape[0], e0bnd[7])
    e0bnd_p[8] = e0_array[8] + (e0_array[8]-e0bnd[7])

    # ;; [9] lower boundary of energy bin
    e0bnd_m = np.insert(e0bnd, 0, e0bnd[0])
    e0bnd_m[0] = e0bnd_m[1]
    e0bnd_m[0] = e0_array[0] - (e0bnd[0]-e0_array[0])

    de_array = e0bnd_p-e0bnd_m  # ;; width of energy bin [9]
    de_reform = np.reshape(de_array, [dim_array[0], 1, 1])
    de_rebin1 = np.repeat(de_reform, dim_array[1],
                         axis=1)  # repeated across spin phase(azimuth)
    dist['denergy'] = np.repeat(de_rebin1, n_times,
                         axis=2)  # repeated across n_times

    dist['n_energy'] = dim_array[0]

    dist['n_bins'] = np.product(dim_array[1:])  #   # phis
   
    #  ;; azimuthal angle in spin direction

    angarr = np.zeros(shape=(2,3))
    angarr[1] = 90.-10.  # ;; these angles should be given as particle flux dirs.
    
    spinper = sc0_dt[tuple([index])]  # ;; spin period [n_times]
    if spinper.size == 1:
        spinper = np.array([spinper])
    rel_sct_time = np.zeros(shape=(n_times, 16))
    rel_sct_time = phtime[tuple([index])] + sctintgt[tuple([index])] / 2.\
                    - np.repeat(np.reshape( phtime[tuple([index]), 0],
                                    (n_times, 1)), 16, axis=1)
    
    phissi = np.array([angarr[1, 1]]) - (90. + 21.6)  # ;; [(1)]
    
    spinph_ofst = (rel_sct_time / np.repeat(spinper.reshape(n_times, 1), 16, axis=1)).T * 360.

    phi0_1_reform = np.reshape(phissi, [1, 1, 1])
    phi0_1_rebin1 = np.repeat(phi0_1_reform, dim_array[1],
                             axis=1)  # repeated across spin phase(azimuth)
    phi0_1_rebin2 = np.repeat(phi0_1_rebin1, dim_array[0],
                             axis=0)  # repeated across energy
    phi0_1 = np.repeat(phi0_1_rebin2,     n_times,
                              axis=2)  # repeated across n_times
    phi0_2_reform = np.reshape(spinph_ofst, [1, dim_array[1], n_times])
    phi0_2 = np.repeat(phi0_2_reform, dim_array[0],
                             axis=0)  # repeated across energy
    phi0 = phi0_1 + phi0_2

    """
    ;;  phi angle for the start of each spin phase
    ;;    + offset angle for each spin phase
    """
    dist['phi'] = np.fmod((phi0 + 360.), 360.)
    dist['dphi'] = np.full(shape=np.insert(dim_array, dim_array.shape[0],
                                           n_times), fill_value=22.5)
    #  ;; 22.5 deg as a constant

    del phi0  # ;; Clean huge arrays

    dist['n_phi'] = dim_array[1]

    #  ;; elevation angle
    elev = np.array([angarr[0, 1]])  # ;; [(1)]
    elev_reform = np.reshape(elev, [1, 1, 1])
    elev_rebin1 = np.repeat(elev_reform, dim_array[1],
                             axis=1)  # repeated across spin phase(azimuth)
    elev_rebin2 = np.repeat(elev_rebin1, dim_array[0],
                             axis=0)  # repeated across energy
    dist['theta'] = np.repeat(elev_rebin2, n_times,
                             axis=2)  # repeated across n_times

    dist['dtheta'] = np.full(shape=np.insert(dim_array, dim_array.shape[0],
                                             n_times), fill_value=20.)  # ;; 20 deg (+/- 10 deg)  as a constant

    dist['n_theta'] = 1

    return dist


import numpy as np

from copy import deepcopy

from pyspedas import tnames, tinterpol
from pytplot import time_double
from pytplot import time_string

from pyspedas.particles.moments.spd_pgs_moments import spd_pgs_moments
from pyspedas.particles.spd_part_products.spd_pgs_regrid import spd_pgs_regrid
from pytplot import get_timespan, get_data, store_data

from .erg_mepe_get_dist import erg_mepe_get_dist
from .erg_mepi_get_dist import erg_mepi_get_dist
from .erg_pgs_clean_data import erg_pgs_clean_data
from .erg_pgs_limit_range import erg_pgs_limit_range
from .erg_convert_flux_units import erg_convert_flux_units
from .erg_pgs_moments_tplot import erg_pgs_moments_tplot
from .erg_pgs_make_fac import erg_pgs_make_fac
from .erg_pgs_make_e_spec import erg_pgs_make_e_spec
from .erg_pgs_make_theta_spec import erg_pgs_make_theta_spec
from .erg_pgs_make_phi_spec import erg_pgs_make_phi_spec
from .erg_pgs_do_fac import erg_pgs_do_fac
from .erg_pgs_progress_update import erg_pgs_progress_update
from .erg_pgs_make_tplot import erg_pgs_make_tplot

def erg_mep_part_products(
    in_tvarname,
    species=None,
    outputs=['energy'],
    no_ang_weighting=True,
    suffix='',
    units='flux',
    datagap=16.1,
    regrid=[32, 16],
    pitch=[0., 180.],
    theta=[-90., 90.],
    phi_in=[0., 360.],
    gyro=[0., 360.],
    energy=None,
    fac_type='mphism',
    trange=None,
    mag_name=None,
    pos_name=None,
    relativistic=False,
    no_regrid=False
    ):

    if len(tnames(in_tvarname)) < 1:
        print('No input data, please specify tplot variable!')
        return 0

    in_tvarname = tnames(in_tvarname)[0]
    instnm = in_tvarname.split('_')[1]  #  ;; mepe or mepi

    if no_ang_weighting:
        no_regrid = True

    if isinstance(outputs, str):
        outputs_lc = outputs.lower()
        outputs_lc = outputs_lc.split(' ')
    elif isinstance(outputs, list):
        outputs_lc = []
        for output in outputs:
            outputs_lc.append(output.lower())

    units_lc = units.lower()

    if phi_in != [0., 360.]:
        if abs(phi_in[1] - phi_in[0]) > 360.:
            print('ERROR: Phi restrictions must have range no larger than 360 deg')
            return 0
        
        phi = phi_in  # Survey or implement of spd_pgs_map_azimuth() have not conducted yet.
        if phi[0] == phi[1]:
            phi = [0., 360.]


    if abs(gyro[1] - gyro[0]) > 360.:
        print('ERROR: Gyro restrictions must have range no larger than 360 deg')
        return 0
    if gyro[0] == gyro[1]:
        gyro = [0., 360.]


    """
    ;;Create energy spectrogram after FAC transformation if limits are not 
    ;;identical to the default.
    """
    if (gyro != [0., 360.]) or (pitch != [0.,180.]):
        idx = np.where(np.array(outputs_lc) == 'energy')[0]
        if (idx.shape[0] > 0) and ('fac_energy' not in outputs_lc):
            idx = idx[0]
            outputs_lc[idx] = 'fac_energy'

        idx = np.where(np.array(outputs_lc) == 'moments')[0]
        if (idx.shape[0] > 0) and ('fac_moments' not in outputs_lc):
            idx = idx[0]
            outputs_lc[idx] = 'fac_moments'

    #  ;;Preserve the original time range
    tr_org = get_timespan(in_tvarname)

    if instnm == 'mepe':
        times_array = erg_mepe_get_dist(in_tvarname, species=species, units=units_lc, time_only=True)
    elif instnm == 'mepi':
        times_array = erg_mepi_get_dist(in_tvarname, species=species, units=units_lc, time_only=True)

    if trange is not None:
        
        trange_double = time_double(trange)
        time_indices = np.where((times_array >= trange_double[0]) \
                                & (times_array <= trange_double[1]))[0]
        
        if time_indices.shape[0] < 1:
            print(f'No ,{in_tvarname}, data for time range ,{time_string(trange_double)}')

    elif trange is None:
        time_indices = np.arange(times_array.shape[0])

    times_array = times_array[time_indices]



    if instnm == 'mepe':
        dist = erg_mepe_get_dist(in_tvarname, 0, species=species, units=units_lc)
    elif instnm == 'mepi':
        dist = erg_mepi_get_dist(in_tvarname, 0, species=species, units=units_lc)

    if 'energy' in outputs_lc:
        out_energy = np.zeros((times_array.shape[0], dist['n_energy']))
        out_energy_y = np.zeros((times_array.shape[0], dist['n_energy']))
    if 'theta' in outputs_lc:
        out_theta = np.zeros((times_array.shape[0], dist['n_theta']))
        out_theta_y = np.zeros((times_array.shape[0], dist['n_theta']))
    if 'phi' in outputs_lc:
        out_phi = np.zeros((times_array.shape[0], dist['n_phi']))
        out_phi_y = np.zeros((times_array.shape[0], dist['n_phi']))

    if 'gyro' in outputs_lc:
        out_gyro = np.zeros((times_array.shape[0], regrid[0]))
        out_gyro_y = np.zeros((times_array.shape[0], regrid[0]))

    if 'pa' in outputs_lc:
        out_pad = np.zeros((times_array.shape[0], regrid[1]))
        out_pad_y = np.zeros((times_array.shape[0], regrid[1]))

    if 'moments' in outputs_lc:
        out_density = np.zeros(times_array.shape[0])
        out_avgtemp = np.zeros(times_array.shape[0])
        out_vthermal = np.zeros(times_array.shape[0])
        out_flux = np.zeros([times_array.shape[0], 3])
        out_velocity = np.zeros([times_array.shape[0], 3])
        out_mftens = np.zeros([times_array.shape[0], 6])
        out_ptens = np.zeros([times_array.shape[0], 6])
        out_ttens = np.zeros([times_array.shape[0], 3, 3])

    if 'fac_energy' in outputs_lc:
        out_fac_energy = np.zeros((times_array.shape[0], dist['n_energy']))
        out_fac_energy_y = np.zeros((times_array.shape[0], dist['n_energy']))

    if 'fac_moments' in outputs_lc:
        out_fac_density = np.zeros(times_array.shape[0])
        out_fac_avgtemp = np.zeros(times_array.shape[0])
        out_fac_vthermal = np.zeros(times_array.shape[0])
        out_fac_flux = np.zeros([times_array.shape[0], 3])
        out_fac_velocity = np.zeros([times_array.shape[0], 3])
        out_fac_mftens = np.zeros([times_array.shape[0], 6])
        out_fac_ptens = np.zeros([times_array.shape[0], 6])
        out_fac_ttens = np.zeros([times_array.shape[0], 3, 3])
    out_vars = []
    last_update_time = None

    """
    ;;--------------------------------------------------------
    ;;Prepare support data
    ;;--------------------------------------------------------
    """
    # ;;create rotation matrix to B-field aligned coordinates if needed
    
    fac_outputs = ['pa','gyro','fac_energy', 'fac_moments']
    fac_requested = len(set(outputs_lc).intersection(fac_outputs)) > 0
    if fac_requested:
        """
        ;; Create magnetic field data with times shifted by half of spin
        ;; periods
        """
        if (mag_name is None) or (len(tnames(mag_name)) < 1):
            print('Cannot find the magnetic field data given by keyword mag_name! EXIT!')
            return
        mag_data = get_data(mag_name)
        dt_array = mag_data[0][1:] - mag_data[0][:-1]
        dt_array = np.insert(dt_array, dt_array.shape[0], dt_array[-1])  # ;; Note that the last value might not be correct.
        mag_name_sftd = mag_name + '_shifted'
        store_data(mag_name_sftd, data={'x':mag_data[0] + dt_array/2.,
                                        'y':mag_data[1]})
        """
        ;;The time shift applied above assumes that the time labels of MGF
        ;;data correspond to the spin start times. Otherwise this should
        ;;be modified properly.
        """

        fac_matrix = erg_pgs_make_fac(times_array, mag_name_sftd, pos_name, fac_type=fac_type)

        if fac_matrix is None:
            # problem creating the FAC matrices
            fac_requested = False

    #  ;;create the magnetic field vector array for moment calculation
    magf = np.array([0., 0., 0.])
    no_mag_for_moments = False

    if ('moments' in outputs_lc) or ('fac_moments' in outputs_lc):

        no_mag = mag_name is None
        magnm = tnames(mag_name)
        if (len(magnm) < 1) or no_mag:
            print('the magnetic field data is not given!')
            no_mag_for_moments = True
        else:
            magnm = magnm[0]

            """
            ;; Create magnetic field data with times shifted by half of spin
            ;; periods
            """

            mag_data = get_data(magnm)
            dt_array = mag_data[0][1:] - mag_data[0][:-1]
            dt_array = np.insert(dt_array, 
            dt_array.shape[0], dt_array[-1])#;; Note that the last value might not be correct.
            magnm_sftd = magnm + '_shifted'
            store_data(magnm_sftd, data={'x':mag_data[0] + dt_array / 2.,
                                         'y':mag_data[1]})
            tinterpol(magnm_sftd, times_array, newname=magnm_sftd)
            magf = get_data(magnm_sftd)[1]  #  ;; [ time, 3] nT

    """
    ;;-------------------------------------------------
    ;; Loop over time to build spectrograms and/or moments
    ;;-------------------------------------------------
    """
    for index in range(time_indices.shape[0]):

        last_update_time = erg_pgs_progress_update(last_update_time=last_update_time,
             current_sample=index, total_samples=time_indices.shape[0], type_string=in_tvarname)

        #  ;; Get the data structure for this sample

        if instnm == 'mepe':

            dist = erg_mepe_get_dist(in_tvarname, time_indices[index], species=species, units=units_lc)

        elif instnm == 'mepi':
            dist = erg_mepi_get_dist(in_tvarname, time_indices[index], species=species, units=units_lc)

        if magf.ndim == 2:
            magvec = magf[index]
        elif magf.ndim == 1:
            magvec = magf

        clean_data = erg_pgs_clean_data(dist, units=units_lc,relativistic=relativistic, magf=magvec)

        if fac_requested:
            pre_limit_bins = deepcopy(clean_data['bins'])

        clean_data = erg_pgs_limit_range(clean_data, phi=phi_in, theta=theta, energy=energy, no_ang_weighting=no_ang_weighting)

        if ('moments' in outputs_lc) or ('fac_moments' in outputs_lc):
            clean_data_eflux = erg_convert_flux_units(clean_data, units='eflux')
            magfarr = deepcopy(magf)
            clean_data_eflux_for_moments = deepcopy(clean_data_eflux)
            clean_data_eflux_for_moments['data'] = np.where(clean_data_eflux_for_moments['bins'] == 0,
                                                            0,clean_data_eflux_for_moments['data'])
            moments = spd_pgs_moments(clean_data_eflux_for_moments)

            if 'moments' in outputs_lc:
                out_density[index] = moments['density']
                out_avgtemp[index] = moments['avgtemp']
                out_vthermal[index] = moments['vthermal']
                out_flux[index, :] = moments['flux']
                out_velocity[index, :] = moments['velocity']
                out_mftens[index, :] = moments['mftens']
                out_ptens[index, :] = moments['ptens']
                out_ttens[index, :] = moments['ttens']

        #  ;;Build theta spectrogram
        if 'theta' in outputs_lc:
            out_theta_y[index, :], out_theta[index, :] = erg_pgs_make_theta_spec(clean_data, resolution=dist['n_theta'],no_ang_weighting=no_ang_weighting)

        #  ;;Build energy spectrogram
        if 'energy' in outputs_lc:
            out_energy_y[index, :], out_energy[index, :] = erg_pgs_make_e_spec(clean_data)

        #  ;;Build phi spectrogram
        if 'phi' in outputs_lc:
            out_phi_y[index, :], out_phi[index, :] = erg_pgs_make_phi_spec(clean_data, resolution=dist['n_phi'],no_ang_weighting=no_ang_weighting)

        #  ;;Perform transformation to FAC, (regrid data), and apply limits in new coords
        
        if fac_requested:
            
            # ;limits will be applied to energy-aligned bins
            clean_data['bins'] = deepcopy(pre_limit_bins)
            clean_data = erg_pgs_limit_range(clean_data, phi=phi_in, theta=theta, energy=energy, no_ang_weighting=no_ang_weighting)

            # ;perform FAC transformation and interpolate onto a new, regular grid 
            clean_data = erg_pgs_do_fac(clean_data, fac_matrix[index, :, :])

            #;nearest neighbor interpolation to regular grid in FAC
            if not no_regrid:
                if (not np.all(np.isnan(clean_data['theta']))) and (not np.all(np.isnan(clean_data['phi']))):
                    clean_data = spd_pgs_regrid(clean_data, regrid)

            clean_data['theta'] = 90.0-clean_data['theta']  #  ;pitch angle is specified in co-latitude

            # ;apply gyro & pitch angle limits(identical to phi & theta, just in new coords)
            clean_data = erg_pgs_limit_range(clean_data, theta=pitch, phi=gyro, no_ang_weighting=no_ang_weighting)

            if 'pa' in outputs_lc:
                # ;Build pitch angle spectrogram
                out_pad_y[index, :], out_pad[index, :] = erg_pgs_make_theta_spec(clean_data, colatitude=True, resolution=regrid[1], no_ang_weighting=no_ang_weighting)

            if 'gyro' in outputs_lc:
                # ;Build gyrophase spectrogram
                out_gyro_y[index, :], out_gyro[index, :] = erg_pgs_make_phi_spec(clean_data, resolution=regrid[0], no_ang_weighting=no_ang_weighting)

            if 'fac_energy' in outputs_lc:
                out_fac_energy_y[index, :], out_fac_energy[index, :] = erg_pgs_make_e_spec(clean_data)

            if 'fac_moments' in outputs_lc:
                clean_data['theta'] = 90. - clean_data['theta'] # ;convert back to latitude for moments calc
                temp_dict = {'charge': dist['charge'],
                             'magf': magvec,
                             'species': dist['species'],
                             'sc_pot': 0.,
                             'units_name': units_lc}
                temp_dict.update(clean_data)
                clean_data = deepcopy(temp_dict)
                del temp_dict
                clean_data_eflux = erg_convert_flux_units(clean_data, units='eflux')
                clean_data_eflux_for_moments = deepcopy(clean_data_eflux)
                clean_data_eflux_for_moments['data'] = np.where(clean_data_eflux_for_moments['bins'] == 0,
                                                            0,clean_data_eflux_for_moments['data'])
                fac_moments = spd_pgs_moments(clean_data_eflux_for_moments)

                out_fac_density[index] = fac_moments['density']
                out_fac_avgtemp[index] = fac_moments['avgtemp']
                out_fac_vthermal[index] = fac_moments['vthermal']
                out_fac_flux[index, :] = fac_moments['flux']
                out_fac_velocity[index, :] = fac_moments['velocity']
                out_fac_mftens[index, :] = fac_moments['mftens']
                out_fac_ptens[index, :] = fac_moments['ptens']
                out_fac_ttens[index, :] = fac_moments['ttens']



    if 'energy' in outputs_lc:
        output_tplot_name = in_tvarname+'_energy' + suffix
        erg_pgs_make_tplot(output_tplot_name, x=times_array, y=out_energy_y, z=out_energy, units=units, ylog=True, ytitle=dist['data_name'] + ' \\ energy (eV)',relativistic=relativistic)
        out_vars.append(output_tplot_name)
    if 'theta' in outputs_lc:
        output_tplot_name = in_tvarname+'_theta' + suffix
        erg_pgs_make_tplot(output_tplot_name, x=times_array, y=out_theta_y, z=out_theta, units=units, ylog=False, ytitle=dist['data_name'] + ' \\ theta (deg)',relativistic=relativistic)
        out_vars.append(output_tplot_name)
    if 'phi' in outputs_lc:
        output_tplot_name = in_tvarname+'_phi' + suffix
        erg_pgs_make_tplot(output_tplot_name, x=times_array, y=out_phi_y, z=out_phi, units=units, ylog=False, ytitle=dist['data_name'] + ' \\ phi (deg)',relativistic=relativistic)
        out_vars.append(output_tplot_name)

    #  ;;Pitch Angle Spectrograms
    if 'pa' in outputs_lc:
        output_tplot_name = in_tvarname+'_pa' + suffix
        erg_pgs_make_tplot(output_tplot_name, x=times_array, y=out_pad_y, z=out_pad, units=units, ylog=False, ytitle=dist['data_name'] + ' \\ PA (deg)',relativistic=relativistic)
        out_vars.append(output_tplot_name)

    if 'gyro' in outputs_lc:
        output_tplot_name = in_tvarname+'_gyro' + suffix
        erg_pgs_make_tplot(output_tplot_name, x=times_array, y=out_gyro_y, z=out_gyro, units=units, ylog=False, ytitle=dist['data_name'] + ' \\ gyro (deg)',relativistic=relativistic)
        out_vars.append(output_tplot_name)


    #  ;Moments Variables
    if 'moments' in outputs_lc:
        moments = {'density': out_density, 
              'flux': out_flux, 
              'mftens': out_mftens, 
              'velocity': out_velocity, 
              'ptens': out_ptens,
              'ttens': out_ttens,
              'vthermal': out_vthermal,
              'avgtemp': out_avgtemp}
        moments_vars = erg_pgs_moments_tplot(moments, x=times_array, prefix=in_tvarname, suffix=suffix)
        out_vars.extend(moments_vars)

    if 'fac_energy' in outputs_lc:

        output_tplot_name = in_tvarname+'_energy_mag' + suffix
        erg_pgs_make_tplot(output_tplot_name, x=times_array, y=out_fac_energy_y, z=out_fac_energy, units=units, ylog=True, ytitle=dist['data_name'] + ' \\ energy (eV)',relativistic=relativistic)
        out_vars.append(output_tplot_name)

    #  ;FAC Moments Variables
    if 'fac_moments' in outputs_lc:
        fac_moments = {'density': out_fac_density, 
              'flux': out_fac_flux, 
              'mftens': out_fac_mftens, 
              'velocity': out_fac_velocity, 
              'ptens': out_fac_ptens,
              'ttens': out_fac_ttens,
              'vthermal': out_fac_vthermal,
              'avgtemp': out_fac_avgtemp}
        fac_mom_suffix = '_mag' + suffix
        fac_moments_vars = erg_pgs_moments_tplot(fac_moments, x=times_array, prefix=in_tvarname, suffix=fac_mom_suffix)
        out_vars.extend(fac_moments_vars)

    return out_vars



import logging
from pyspedas.mms.mms_config import CONFIG


# the following decorator prints the loaded tplot variables after each load routine call
def print_vars(func):
    def wrapper(*args, **kwargs):
        variables = func(*args, **kwargs)
        if variables is None:
            return None
        if kwargs.get('available') or CONFIG['download_only']:
            logging.info('Available files:')
        else:
            logging.info('Loaded variables:')
        for var in variables:
            logging.info(var)
        return variables
    wrapper.__name__ = func.__name__
    wrapper.__doc__ = func.__doc__
    return wrapper




import math
import numpy as np
from scipy.ndimage.interpolation import shift
from copy import deepcopy

# use nansum from bottleneck if it's installed, otherwise use the numpy one
try:
    import bottleneck as bn
    nansum = bn.nansum
except ImportError:
    nansum = np.nansum

def erg_pgs_make_theta_spec(data_in, resolution=None, colatitude=False, no_ang_weighting=False):
    """
    Builds theta (latitudinal) spectrogram from simplified particle data structure.

    Input:
        data_in: dict
            Particle data structure

    Parameters:
        resolution: int
            Number of theta points to include in the output

        colatitude: bool
            Flag to specify that data is in co-latitude (0, 180); if this is
            set to False (default), the data are assumed to be (-90, 90)

    Returns:
        Tuple containing: (theta values for y-axis, spectrogram values)

    """

    dr = math.pi/180.

    data = deepcopy(data_in)

    # zero inactive bins to ensure areas with no data are represented as NaN
    data['data'] = np.where(data['bins'] == 0,np.nan,data['data'])
    # get number of theta values
    if resolution is None:
        n_theta = len(np.unique(data['theta']))
    else:
        n_theta = resolution

    if colatitude:
        theta_range = [0, 180]
    else:
        theta_range = [-90, 90]

    theta_grid = np.linspace(theta_range[0], theta_range[1], n_theta+1)

    # init this sample's piece of the spectrogram
    ave = np.zeros(n_theta)

    theta_min = data['theta'] - 0.5*data['dtheta']
    theta_max = data['theta'] + 0.5*data['dtheta']
    theta_array = deepcopy(data['theta'])

    # loop over output grid to sum all active data and bin flags
    for i in range(0, n_theta):
       
        if not no_ang_weighting:
           
            weight = np.zeros(theta_min.shape)

            # data bins whose maximum overlaps the current spectrogram bin
            idx_max = np.argwhere((theta_max > theta_grid[i]) & (theta_max < theta_grid[i+1]))
            if idx_max.size != 0:
                weight[idx_max[:, 0].tolist(), idx_max[:, 1].tolist()] = (np.sin(dr * theta_max[idx_max[:, 0].tolist(), idx_max[:, 1].tolist()]) - np.sin(dr * theta_grid[i])) * data['dphi'][idx_max[:, 0].tolist(), idx_max[:, 1].tolist()]

            # data bins whose minimum overlaps the current spectrogram bin
            idx_min = np.argwhere((theta_min > theta_grid[i]) & (theta_min < theta_grid[i+1]))
            if idx_min.size != 0:
                weight[idx_min[:, 0].tolist(), idx_min[:, 1].tolist()] = (np.sin(dr * theta_grid[i+1]) - np.sin(dr * theta_min[idx_min[:, 0].tolist(), idx_min[:, 1].tolist()])) * data['dphi'][idx_min[:, 0].tolist(), idx_min[:, 1].tolist()]

            # data bins contained within the current spectrogram bin
            max_set = set([tuple(m) for m in idx_max])
            min_set = set([tuple(m) for m in idx_min])
            contained = np.array([m for m in max_set & min_set])
            if contained.size != 0:
                weight[contained[:, 0].tolist(), contained[:, 1].tolist()] = (np.sin(dr * theta_max[contained[:, 0].tolist(), contained[:, 1].tolist()]) - np.sin(dr * theta_min[contained[:, 0].tolist(), contained[:, 1].tolist()])) * data['dphi'][contained[:, 0].tolist(), contained[:, 1].tolist()]

            # data bins that completely cover the current spectrogram bin
            idx_all = np.argwhere((theta_min <= theta_grid[i]) & (theta_max >= theta_grid[i+1]))
            if idx_all.size != 0:
                weight[idx_all[:, 0].tolist(), idx_all[:, 1].tolist()] = (np.sin(dr * theta_grid[i+1]) - np.sin(dr * theta_grid[i])) * data['dphi'][idx_all[:, 0].tolist(), idx_all[:, 1].tolist()]

            # combine indices
            idx = np.concatenate((idx_min, idx_max, idx_all))

            if idx_max.size + idx_min.size + idx_all.size > 0:
                # normalize weighting to selected, active bins
                weight[idx[:, 0].tolist(), idx[:, 1].tolist()] = weight[idx[:, 0].tolist(), idx[:, 1].tolist()] * data['bins'][idx[:, 0].tolist(), idx[:, 1].tolist()]\
                                                                * np.isfinite(data['data'][idx[:, 0].tolist(), idx[:, 1].tolist()])
                weight = weight/nansum(weight)

                # average
                ave[i] = nansum(data['data'][idx[:, 0].tolist(), idx[:, 1].tolist()]*weight[idx[:, 0].tolist(), idx[:, 1].tolist()])
       
        else: # ;;without weighting by dtheta and dphi
            id_array = np.argwhere(np.isfinite(data['data'])\
                     & (data['bins'] == 1)\
                     & (theta_array > theta_grid[i])\
                     & (theta_array < theta_grid[i+1]))

            if len(id_array) > 0:
                ave[i] = np.nanmean(data['data'][id_array[:, 0].tolist(), id_array[:, 1].tolist()])

    # get y axis
    y = (theta_grid+shift(theta_grid, 1))/2.0
    y = y[1:]

    return (y, ave)




"""
Transform datetimes from decimal to string.

Examples
--------
time_string()
time_string([1498780799.1234, 1512087321.1234])

Notes
-----
Compare to https://www.epochconverter.com/

"""
import logging
import pytplot

def time_string_one(float_time=None, fmt=None):
    """
    Transform a single float daytime value to string.

    Parameters
    ----------
    float_time : float, optional
        Input time.
        The default is None, which returns the time now.
    fmt : float, optional
        Time format.
        The default is None, which uses '%Y-%m-%d %H:%M:%S.%f'.

    Returns
    -------
    str
        Datetime as string.

    """
    logging.info("time_string_one has been moved to the pytplot package. Please update your imports!")
    logging.info("This version will eventually be removed.")
    return pytplot.time_string_one(float_time=float_time,fmt=fmt)

def time_string(float_time=None, fmt=None):
    """
    Transform a list of float daytime values to a list of strings.

    Parameters
    ----------
    float_time: float/list of floats, optional
        Input time.
        The default is None, which returns the time now.
    fmt: str, optional
        Time format.
        The default is None, which uses '%Y-%m-%d %H:%M:%S.%f'.

    Returns
    -------
    list of str
        Datetimes as string.

    """
    logging.info("time_string has been moved to the pytplot package. Please update your imports!")
    logging.info("This version will eventually be removed.")
    return pytplot.time_string(float_time=float_time,fmt=fmt)

def time_datetime(time=None, tz=None):
    """Find python datetime.

    Transform a list of float daytime values to a list of pythonic
        'datetime.datetime' values.

    Parameters
    ----------
    time: float/list of floats or str/list of str, optional
        Input time.
        The default is None, which returns the time now.

    Returns
    -------
    list of datetime.datetime
        Datetimes as `datetime.datetime`.

    """
    logging.info("time_datetime has been moved to the pytplot package. Please update your imports!")
    logging.info("This version will eventually be removed.")
    return pytplot.time_datetime(time=time,tz=tz)



import numpy as np


def slice2d_nearest(dists, time, samples):
    """
    Get a time range that encompasses a specified number of
    samples closest to a specified time.

    Input
    ------
        dists: list
            List containing distribution data structures

        time: float
            Unix time

        samples: int
            Number of samples to return

    Returns
    --------
        List containing [start time, end time] for the time range
    """
    start_times = np.zeros(len(dists))
    end_times = np.zeros(len(dists))

    for i, dist in enumerate(dists):
        start_times[i] = dist['start_time']
        end_times[i] = dist['end_time']

    # use center
    distance = np.abs((start_times + end_times)/2.0 - time)

    # get indices for N closest samples
    idx = np.argsort(distance)
    if samples > len(idx):
        n = len(idx)
    else:
        n = samples
    idx = idx[0:n]

    # full time range
    return [np.nanmin(start_times[idx]),
            np.nanmax(end_times[idx])]



import cdflib
import numpy as np
from pyspedas import tnames
from pytplot import time_float
from pytplot import clip, get_data, options, store_data, ylim, zlim

from ..load import load


def pwe_wfc(trange=['2017-04-01/12:00:00', '2017-04-01/13:00:00'],
            datatype='waveform',
            mode='65khz',
            level='l2',
            suffix='',
            coord='sgi',
            component='all',
            get_support_data=False,
            varformat=None,
            varnames=[],
            downloadonly=False,
            notplot=False,
            no_update=False,
            uname=None,
            passwd=None,
            time_clip=False,
            ror=True):
    """
    This function loads data from the PWE experiment from the Arase mission

    Parameters:
        trange : list of str
            time range of interest [starttime, endtime] with the format
            'YYYY-MM-DD','YYYY-MM-DD'] or to specify more or less than a day
            ['YYYY-MM-DD/hh:mm:ss','YYYY-MM-DD/hh:mm:ss']

        datatype: str
            Data type; Valid options:

        level: str
            Data level; Valid options:

        suffix: str
            The tplot variable names will be given this suffix.  By default,
            no suffix is added.

        get_support_data: bool
            Data with an attribute "VAR_TYPE" with a value of "support_data"
            will be loaded into tplot.  By default, only loads in data with a
            "VAR_TYPE" attribute of "data".

        varformat: str
            The file variable formats to load into tplot.  Wildcard character
            "*" is accepted.  By default, all variables are loaded in.

        varnames: list of str
            List of variable names to load (if not specified,
            all data variables are loaded)

        downloadonly: bool
            Set this flag to download the CDF files, but not load them into
            tplot variables

        notplot: bool
            Return the data in hash tables instead of creating tplot variables

        no_update: bool
            If set, only load data from your local cache

        time_clip: bool
            Time clip the variables to exactly the range specified in the trange keyword

        ror: bool
            If set, print PI info and rules of the road

    Returns:
        List of tplot variables created.

    """
    initial_notplot_flag = False
    if notplot:
        initial_notplot_flag = True

    file_res = 3600.

    if level == 'l2':
        prefix = 'erg_pwe_wfc_'+level+'_' + mode + '_'

    loaded_data = []
    if level == 'l2':
        if datatype == 'waveform':
            tplot_name_list = []
            if component == 'all':
                component_list = ['e', 'b']
            elif (component == 'e') or (component == 'b'):
                component_list = [component]
            for com in component_list:
                prefix = 'erg_pwe_wfc_' + level + '_' + com + '_' + mode + '_'
                pathformat = 'satellite/erg/pwe/wfc/'+level+'/'+datatype+'/%Y/%m/erg_pwe_wfc_' + \
                    level+'_'+com+'_'+datatype+'_'+mode+'_'+coord+'_%Y%m%d%H_v??_??.cdf'
                loaded_data.append(load(pathformat=pathformat, trange=trange, level=level, datatype=datatype, file_res=file_res, prefix=prefix, suffix=suffix, get_support_data=get_support_data,
                                   varformat=varformat, varnames=varnames, downloadonly=downloadonly, notplot=notplot, time_clip=time_clip, no_update=no_update, uname=uname, passwd=passwd))
                if com == 'e':
                    tplot_name_list += [prefix +
                                        'Ex_waveform', prefix + 'Ey_waveform']
                elif com == 'b':
                    tplot_name_list += [prefix + 'Bx_waveform',
                                        prefix + 'By_waveform', prefix + 'Bz_waveform']
        elif datatype == 'spec':
            prefix_list = []
            component_suffix_list = []
            if component == 'all':
                component_list = ['e', 'b']
            elif (component == 'e') or (component == 'b'):
                component_list = [component]
            for com in component_list:
                prefix = 'erg_pwe_wfc_' + level + '_' + com + '_' + mode + '_'
                pathformat = 'satellite/erg/pwe/wfc/'+level+'/'+datatype+'/%Y/%m/erg_pwe_wfc_' + \
                    level+'_'+com+'_'+datatype+'_'+mode+'_%Y%m%d%H_v??_??.cdf'
                loaded_data.append(load(pathformat=pathformat, trange=trange, level=level, datatype=datatype, file_res=file_res, prefix=prefix, suffix=suffix, get_support_data=get_support_data,
                                   varformat=varformat, varnames=varnames, downloadonly=downloadonly, notplot=notplot, time_clip=time_clip, no_update=no_update, uname=uname, passwd=passwd))
                prefix_list.append(prefix)
                component_suffix_list.append(com.upper() + '_spectra')

    if (len(loaded_data) > 0) and ror:

        try:
            if isinstance(loaded_data, list):
                if downloadonly:
                    cdf_file = cdflib.CDF(loaded_data[-1][-1])
                    gatt = cdf_file.globalattsget()
                elif notplot:
                    gatt = loaded_data[-1][list(loaded_data[-1].keys())[-1]]['CDF']['GATT']
                else:
                    gatt = get_data(loaded_data[-1][-1], metadata=True)['CDF']['GATT']


            # --- print PI info and rules of the road
            print(' ')
            print(' ')
            print(
                '**************************************************************************')
            print(gatt["LOGICAL_SOURCE_DESCRIPTION"])
            print('')
            print('Information about ERG PWE WFC')
            print('')
            print('PI: ', gatt['PI_NAME'])
            print("Affiliation: "+gatt["PI_AFFILIATION"])
            print('')
            print('RoR of ERG project common: https://ergsc.isee.nagoya-u.ac.jp/data_info/rules_of_the_road.shtml.en')
            print(
                'RoR of PWE/WFC: https://ergsc.isee.nagoya-u.ac.jp/mw/index.php/ErgSat/Pwe/Wfc')
            print('')
            print('Contact: erg_pwe_info at isee.nagoya-u.ac.jp')
            print(
                '**************************************************************************')
        except:
            print('printing PI info and rules of the road was failed')

    if initial_notplot_flag or downloadonly:
        return loaded_data

    if datatype == 'spec':
        trange_in_float = time_float(trange)
        for i in range(len(prefix_list)):
            t_plot_name = prefix_list[i] + component_suffix_list[i]
            options(t_plot_name, 'spec', 1)
            options(t_plot_name, 'colormap', 'jet')
            options(t_plot_name, 'ylog', 1)
            options(t_plot_name, 'zlog', 1)
            options(t_plot_name, 'ysubtitle', '[Hz]')
            ylim(t_plot_name, 32., 2e4)
            if 'E_spectra' in component_suffix_list[i]:
                zlim(t_plot_name, 1e-9, 1e-2)
                options(t_plot_name, 'ztitle', '[mV^2/m^2/Hz]')
                options(t_plot_name, 'ytitle', 'E\nspectra')
            elif 'B_spectra' in component_suffix_list[i]:
                zlim(t_plot_name, 1e-4, 1e2)
                options(t_plot_name, 'ztitle', '[pT^2/Hz]')
                options(t_plot_name, 'ytitle', 'B\nspectra')

            get_data_vars = get_data(t_plot_name)
            time_array = get_data_vars[0]
            if time_array[0] <= trange_in_float[0]:
                t_ge_indices = np.where(time_array <= trange_in_float[0])
                t_min_index = t_ge_indices[0][-1]
            else:
                t_min_index = 0
            if trange_in_float[1] <= time_array[-1]:
                t_le_indices = np.where(trange_in_float[1] <= time_array)
                t_max_index = t_le_indices[0][0]
            else:
                t_max_index = -1
            if t_min_index == t_max_index:
                t_max_index = + 1
            if (t_min_index != 0) or (t_max_index != -1):
                meta_data = get_data(t_plot_name, metadata=True)
                store_data(t_plot_name, newname=t_plot_name +
                           '_all_loaded_time_range')
                store_data(t_plot_name, data={'x': time_array[t_min_index:t_max_index],
                                              'y': get_data_vars[1][t_min_index:t_max_index],
                                              'v': get_data_vars[2]},
                           attr_dict=meta_data)
                options(t_plot_name, 'zlog', 1)
                if 'E_spectra' in t_plot_name:
                    zlim(t_plot_name,  1e-9, 1e-2)
                elif 'B_spectra' in t_plot_name:
                    zlim(t_plot_name,  1e-4, 1e2)

    if datatype == 'waveform':
        trange_in_float = time_float(trange)
        yn = ''
        all_time_range_flag = False
        if trange_in_float[1] - trange_in_float[0] <= 0.:
            yn = input('Invalid time range. Use full time range ?:[y/n] ')
            if yn == 'y':
                all_time_range_flag = True
                t_min_index = 0
                t_max_index = -1
            else:
                return
        for t_plot_name in tplot_name_list:
            get_data_vars = get_data(t_plot_name)
            dl_in = get_data(t_plot_name, metadata=True)
            time_array = get_data_vars[0]
            if not all_time_range_flag:
                if time_array[0] <= trange_in_float[0]:
                    t_ge_indices = np.where(time_array <= trange_in_float[0])
                    t_min_index = t_ge_indices[0][-1]
                else:
                    t_min_index = 0
                if trange_in_float[1] <= time_array[-1]:
                    t_le_indices = np.where(trange_in_float[1] <= time_array)
                    t_max_index = t_le_indices[0][0]
                else:
                    t_max_index = -1
                if t_min_index == t_max_index:
                    t_max_index = + 1
            data = np.where(get_data_vars[1] <= -1e+30, np.nan, get_data_vars[1])
            dt = get_data_vars[2]
            ndt = dt.size
            ndata = (t_max_index - t_min_index) * ndt
            time_new = (np.tile(
                time_array[t_min_index:t_max_index], (ndt, 1)).T + dt * 1e-3).reshape(ndata)
            data_new = data[t_min_index:t_max_index].reshape(ndata)
            store_data(t_plot_name, data={
                       'x': time_new, 'y': data_new}, attr_dict=dl_in)
            options(t_plot_name, 'ytitle', '\n'.join(t_plot_name.split('_')))
            # ylim settings because pytplot.timespan() doesn't affect in ylim.
            # May be it will be no need in future.
            if not all_time_range_flag:
                if time_new[0] <= trange_in_float[0]:
                    t_min_index = np.where(
                        (time_new <= trange_in_float[0]))[0][-1]
                else:
                    t_min_index = 0
                if trange_in_float[1] <= time_new[-1]:
                    t_max_index = np.where(
                        trange_in_float[1] <= time_new)[0][0]
                else:
                    t_max_index = -1
            ylim_min = np.nanmin(data_new[t_min_index:t_max_index])
            ylim_max = np.nanmax(data_new[t_min_index:t_max_index])
            ylim(t_plot_name, ylim_min, ylim_max)

    return loaded_data



"""
This module provides functions for transforming MMS vector fields from one coordinate system to another using quaternion rotation.

To use this module, you will need to install the SpacePy package: pip install spacepy.

The main function of this module is mms_cotrans_qrotate, which performs a quaternion rotation on a PyTplot variable. The function takes in the names of the input and quaternion tplot variables, the name of the output tplot variable, and the coordinate system for the output data. An optional inverse flag allows the user to use the quaternion conjugate on the quaternion data prior to rotating. If the data and quaternion tplot variables are not the same length, the function will interpolate the data to the quaternion timestamps.
"""
import logging
from pytplot import get_data, store_data
from pyspedas import cotrans_set_coord, tinterpol

try:
    import spacepy.coordinates as coord
except ImportError:
    logging.error('SpacePy must be installed to use this module.')
    logging.error('Please install it using: pip install spacepy')


def mms_cotrans_qrotate(in_name, q_name, out_name, out_coord, inverse=False):
    """
    Performs a quaternion rotation on a tplot variable.

    Note: this routine will interpolate the input data to the quaternion
    timestamps if the data aren't the same length as the quaternions

    Parameters
    ----------
        in_name: str
            Tplot variable to be transformed

        q_name: str
            Tplot variable containing the quaternions

        out_name: str
            Tplot variable to be created

        inverse: bool:
            Flag to use the quaternionConjugate on the quaternion data prior to rotating
    """
    data = get_data(in_name)
    metadata = get_data(in_name, metadata=True)

    q_data = get_data(q_name)

    if data is None:
        logging.error(f'Problem reading input tplot variable: {in_name}')
        return

    if q_data is None:
        logging.error(f'Problem reading quaternion variable: {q_name}')
        return

    if len(data.times) != len(q_data.times):
        logging.info('Interpolating the data to the MEC quaternion time stamps.')
        tinterpol(in_name, q_name)
        data = get_data(in_name + '-itrp')

    if inverse:
        quaternion = coord.quaternionConjugate(q_data.y)
    else:
        quaternion = q_data.y

    out_data = coord.quaternionRotateVector(quaternion, data.y)

    saved = store_data(out_name, data={'x': data.times, 'y': out_data}, attr_dict=metadata)
    if saved:
        cotrans_set_coord(out_name, out_coord)




import logging
from pytplot import get_data, store_data, tplot_copy
import numpy as np
from pytplot import data_exists


def apply_oneaxis_correction(rawvar: str,
                             deltavar: str,
                             corrvar: str):
    """
    This function applies spin axis corrections for a single spin axis quantity (RA or DEC).

    Parameters:
        rawvar: str
            Name of tplot variable holding the uncorrected quantity of interest.

        deltavar: str
            Name of tplot variable holding the corrections for the quantity of interest

        corrvar: str
            Name of tplot variable to receive the corrected quantity
    """

    if (data_exists(rawvar)) and (data_exists(deltavar)):
        logging.debug("Applying spin axis corrections")
        res1 = get_data(rawvar)
        raw_times = res1.times
        raw_data = res1.y
        metadata = get_data(rawvar, metadata=True)

        res2 = get_data(deltavar)
        corr_times = res2.times
        corr_data = res2.y

        # Interpolate corrections using input_times

        # In IDL, there was a special case, if the correction had only a single quantity, because the
        # interpol routine would crash with only a single value given. np.interp does the right thing
        # by default.

        interp_correction = np.interp(raw_times, corr_times, corr_data)

        # Apply corrections
        fix_raw = raw_data - interp_correction

        # Make output tplot variable
        store_data(corrvar, data={'x': raw_times, 'y': fix_raw}, attr_dict=metadata)
    elif not (data_exists(deltavar)):
        # If the corrections aren't present, just copy the original data
        logging.info('Spin axis corrections variable ' + deltavar + ' not found, copying ' + rawvar + ' to ' + corrvar)
        tplot_copy(rawvar,corrvar)
    else:
        # Raw variable doesn't exist, nothing to do here.
        logging.info('Spin axis variable ' + rawvar + ' not found, skipping ' + corrvar)


def apply_spinaxis_corrections(spinras: str,
                               spindec: str,
                               delta_spinras: str,
                               delta_spindec: str,
                               corrected_spinras: str,
                               corrected_spindec: str):
    """
    This function applies V03 state spin axis corrections (if present) to state spinras and spindec
    quantities, creating new tplot variables.

    Parameters:
        spinras: str
            Name of tplot variable holding the uncorrected spin axis right ascension.

        spindec: str
            Name of tplot variable holding the uncorrected spin axis declination.

        delta_spinras: str
            Name of tplot variable holding the correction to the spin axis right ascension.

        delta_spindec: str
            Name of tplot variable holding the corrections to the spin axis declination.

        corrected_spinras: str
            Name of tplot variable to receive the corrected spin axis right ascension.

        corrected_spindec: str
            Name of tplot variable to receive the corrected spin axis declination.
    """
    apply_oneaxis_correction(rawvar=spinras, deltavar=delta_spinras, corrvar=corrected_spinras)
    apply_oneaxis_correction(rawvar=spindec, deltavar=delta_spindec, corrvar=corrected_spindec)



from pyspedas.utilities.dailynames import dailynames
from pyspedas.utilities.download import download
from pytplot import time_clip as tclip
from pytplot import cdf_to_tplot

from .config import CONFIG


def load_orbit(trange=['2013-11-5', '2013-11-6'],
               probe='15',
               prefix='',
               suffix='',
               get_support_data=False,
               varformat=None,
               varnames=[],
               downloadonly=False,
               notplot=False,
               no_update=False,
               time_clip=True):
    """
    This function loads GOES orbit data from SPDF:

    https://spdf.gsfc.nasa.gov/pub/data/goes/goes#/orbit/YYYY/

    Parameters
    ----------
        trange : list of str
            time range of interest [starttime, endtime] with the format
            'YYYY-MM-DD','YYYY-MM-DD'] or to specify more or less than a day
            ['YYYY-MM-DD/hh:mm:ss','YYYY-MM-DD/hh:mm:ss']

        probe: str/int or list of strs/ints
            GOES spacecraft #, e.g., probe=15

        prefix: str
            The tplot variable names will be given this prefix.
            By default, no prefix is added.
            If 'probename' then the name will be used, for example g16.

        suffix: str
            The tplot variable names will be given this suffix.
            By default, no suffix is added.

        downloadonly: bool
            Set this flag to download the CDF files, but not load them into
            tplot variables

        notplot: bool
            If set, load the data into dictionaries containing the numpy objects instead
            of creating tplot variables

        no_update: bool
            If set, only load data from your local cache

        time_clip: bool
            Time clip the variables to exactly the range specified in the trange keyword

    Returns
    ----------
        List of tplot variables created. Or list of filenames downloaded.

    """
    remote_data_dir = 'https://spdf.gsfc.nasa.gov/pub/data/goes/'
    out_files = []  # list of local files downloaded
    tvars = []  # list of tplot variables created

    if not isinstance(probe, list):
        probe = [probe]

    for prb in probe:

        # yearly files
        pathformat = 'goes' + str(prb) + '/orbit/%Y/goes' + str(prb) + '_ephemeris_ssc_%Y0101_v??.cdf'

        # find the full remote path names using the trange
        remote_names = dailynames(file_format=pathformat, trange=trange)

        files = download(remote_file=remote_names, remote_path=remote_data_dir,
                         local_path=CONFIG['local_data_dir'], no_download=no_update)

        out_files_local = []

        if files is not None:
            for file in files:
                out_files_local.append(file)

        out_files.extend(out_files_local)

        tvars_local = []
        if not downloadonly:
            if prefix == 'probename':
                prefix_local = 'g' + str(prb) + '_'
            else:
                prefix_local = prefix

            tvars_local = cdf_to_tplot(out_files_local, prefix=prefix_local, suffix=suffix, get_support_data=get_support_data,
                                       varformat=varformat, varnames=varnames, notplot=notplot)
            tvars.extend(tvars_local)

        if time_clip:
            for new_var in tvars_local:
                tclip(new_var, trange[0], trange[1], suffix='')

    if downloadonly:
        return out_files

    return tvars
